"""
Contains utility functions for working with nested python data structures.

A *pytree* is Python nested data structure. It is a tree in the sense that
nodes are Python collections (e.g., list, tuple, dict) and the leaves are
Python values. Furthermore, a pytree should not contain reference cycles.

pytrees are useful for working with nested collections of Tensors. For example,
one can use `tree_map` to map a function over all Tensors inside some nested
collection of Tensors and `tree_leaves` to get a flat list of all Tensors
inside some nested collection. pytrees are helpful for implementing nested
collection support for PyTorch APIs.

This pytree implementation is not very performant due to Python overhead
To improve the performance we can move parts of the implementation to C++.
"""

import dataclasses
import functools
import importlib
import json
import sys
import threading
import types
import warnings
from collections import defaultdict, deque, namedtuple, OrderedDict
from typing import (
    Any,
    Callable,
    cast,
    DefaultDict,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    Hashable,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    OrderedDict as GenericOrderedDict,
    overload,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from typing_extensions import deprecated


__all__ = [
    "PyTree",
    "Context",
    "FlattenFunc",
    "UnflattenFunc",
    "DumpableContext",
    "ToDumpableContextFn",
    "FromDumpableContextFn",
    "TreeSpec",
    "LeafSpec",
    "keystr",
    "key_get",
    "register_pytree_node",
    "tree_flatten",
    "tree_flatten_with_path",
    "tree_unflatten",
    "tree_iter",
    "tree_leaves",
    "tree_leaves_with_path",
    "tree_structure",
    "tree_map",
    "tree_map_with_path",
    "tree_map_",
    "tree_map_only",
    "tree_map_only_",
    "tree_all",
    "tree_any",
    "tree_all_only",
    "tree_any_only",
    "treespec_dumps",
    "treespec_loads",
    "treespec_pprint",
]


T = TypeVar("T")
S = TypeVar("S")
U = TypeVar("U")
R = TypeVar("R")


DEFAULT_TREESPEC_SERIALIZATION_PROTOCOL = 1
NO_SERIALIZED_TYPE_NAME_FOUND = "NO_SERIALIZED_TYPE_NAME_FOUND"


class KeyEntry(Protocol):
    def __hash__(self) -> int:
        ...

    def __eq__(self, other: object) -> bool:
        ...

    def __str__(self) -> str:
        ...

    def get(self, parent: Any) -> Any:
        ...


Context = Any
PyTree = Any
FlattenFunc = Callable[[PyTree], Tuple[List[Any], Context]]
UnflattenFunc = Callable[[Iterable[Any], Context], PyTree]
DumpableContext = Any  # Any json dumpable text
ToDumpableContextFn = Callable[[Context], DumpableContext]
FromDumpableContextFn = Callable[[DumpableContext], Context]
ToStrFunc = Callable[["TreeSpec", List[str]], str]
MaybeFromStrFunc = Callable[[str], Optional[Tuple[Any, Context, str]]]
KeyPath = Tuple[KeyEntry, ...]
FlattenWithKeysFunc = Callable[[PyTree], Tuple[List[Tuple[KeyEntry, Any]], Any]]


# A NodeDef holds two callables:
# - flatten_fn should take the collection and return a flat list of values.
#   It can also return some context that is used in reconstructing the
#   collection.
# - unflatten_fn should take a flat list of values and some context
#   (returned by flatten_fn). It returns the collection by reconstructing
#   it from the list and the context.
# - flatten_with_keys_fn, which is a callable that takes a
#   pytree and returns a list of (keypath, value) pairs and a context.
class NodeDef(NamedTuple):
    type: Type[Any]
    flatten_fn: FlattenFunc
    unflatten_fn: UnflattenFunc
    flatten_with_keys_fn: Optional[FlattenWithKeysFunc]


_NODE_REGISTRY_LOCK = threading.Lock()
SUPPORTED_NODES: Dict[Type[Any], NodeDef] = {}


# _SerializeNodeDef holds the following:
# - typ: the type of the node (e.g., "Dict", "List", etc)
# - serialized_type_name: the fully qualified name of the type, e.g. "collections.OrderedDict"
# - to_dumpable_context takes a TreeSpec, and returns a serialized string format of the
#   context, and the version number
# - from_dumpable_context takes in a string representation of the context, and the
#   version, and returns the deserialized context
class _SerializeNodeDef(NamedTuple):
    typ: Type[Any]
    serialized_type_name: str
    to_dumpable_context: Optional[ToDumpableContextFn]
    from_dumpable_context: Optional[FromDumpableContextFn]


SUPPORTED_SERIALIZED_TYPES: Dict[Type[Any], _SerializeNodeDef] = {}
SERIALIZED_TYPE_TO_PYTHON_TYPE: Dict[str, Type[Any]] = {}


def register_pytree_node(
    cls: Type[Any],
    flatten_fn: FlattenFunc,
    unflatten_fn: UnflattenFunc,
    *,
    serialized_type_name: Optional[str] = None,
    to_dumpable_context: Optional[ToDumpableContextFn] = None,
    from_dumpable_context: Optional[FromDumpableContextFn] = None,
    flatten_with_keys_fn: Optional[FlattenWithKeysFunc] = None,
) -> None:
    """Register a container-like type as pytree node.

    Args:
        cls: the type to register
        flatten_fn: A callable that takes a pytree and returns a flattened
            representation of the pytree and additional context to represent the
            flattened pytree.
        unflatten_fn: A callable that takes a flattened version of the pytree,
            additional context, and returns an unflattened pytree.
        serialized_type_name: A keyword argument used to specify the fully qualified
            name used when serializing the tree spec.
        to_dumpable_context: An optional keyword argument to custom specify how
            to convert the context of the pytree to a custom json dumpable
            representation. This is used for json serialization, which is being
            used in torch.export right now.
        from_dumpable_context: An optional keyword argument to custom specify how
            to convert the custom json dumpable representation of the context
            back to the original context. This is used for json deserialization,
            which is being used in torch.export right now.
        flatten_with_keys_fn: An optional keyword argument to specify how to
            access each pytree leaf's keypath when flattening and tree-mapping.
            Like ``flatten_fn``, but in place of a List[leaf], it should return
            a List[(keypath, leaf)].
    """
    with _NODE_REGISTRY_LOCK:
        if cls in SUPPORTED_NODES:
            raise ValueError(f"{cls} is already registered as pytree node.")

    _private_register_pytree_node(
        cls,
        flatten_fn,
        unflatten_fn,
        serialized_type_name=serialized_type_name,
        to_dumpable_context=to_dumpable_context,
        from_dumpable_context=from_dumpable_context,
        flatten_with_keys_fn=flatten_with_keys_fn,
    )

    try:
        from . import _cxx_pytree as cxx
    except ImportError:
        pass
    else:
        cxx._private_register_pytree_node(
            cls,
            flatten_fn,
            unflatten_fn,
            serialized_type_name=serialized_type_name,
            to_dumpable_context=to_dumpable_context,
            from_dumpable_context=from_dumpable_context,
        )


def _register_namedtuple(
    cls: Type[Any],
    *,
    serialized_type_name: str,
) -> None:
    """
    Registers a namedtuple as a valid pytree node. By default namedtuples are
    valid pytree nodes, but they are not serializable. This API provides the
    argument `serialized_type_name` which allows these namedtuples to be
    serialized.

    Args:
        cls: the dataclass type to register
        serialized_type_name: The serialized name for the dataclass. This is
        required if you want to serialize the pytree TreeSpec containing this
        namedtuple.
    """
    _private_register_pytree_node(
        cls,
        _namedtuple_flatten,
        _namedtuple_unflatten,
        serialized_type_name=serialized_type_name,
        to_dumpable_context=_namedtuple_serialize,
        from_dumpable_context=_namedtuple_deserialize,
        flatten_with_keys_fn=_namedtuple_flatten_with_keys,
    )


@deprecated(
    "`torch.utils._pytree._register_pytree_node` is deprecated. "
    "Please use `torch.utils._pytree.register_pytree_node` instead.",
    category=FutureWarning,
)
def _register_pytree_node(
    cls: Type[Any],
    flatten_fn: FlattenFunc,
    unflatten_fn: UnflattenFunc,
    to_str_fn: Optional[ToStrFunc] = None,  # deprecated
    maybe_from_str_fn: Optional[MaybeFromStrFunc] = None,  # deprecated
    *,
    serialized_type_name: Optional[str] = None,
    to_dumpable_context: Optional[ToDumpableContextFn] = None,
    from_dumpable_context: Optional[FromDumpableContextFn] = None,
    flatten_with_keys_fn: Optional[FlattenWithKeysFunc] = None,
) -> None:
    """Register a container-like type as pytree node for the Python pytree only.

    Args:
        cls: the type to register
        flatten_fn: A callable that takes a pytree and returns a flattened
            representation of the pytree and additional context to represent the
            flattened pytree.
        unflatten_fn: A callable that takes a flattened version of the pytree,
            additional context, and returns an unflattened pytree.
        serialized_type_name: A keyword argument used to specify the fully qualified
            name used when serializing the tree spec.
        to_dumpable_context: An optional keyword argument to custom specify how
            to convert the context of the pytree to a custom json dumpable
            representation. This is used for json serialization, which is being
            used in torch.export right now.
        from_dumpable_context: An optional keyword argument to custom specify how
            to convert the custom json dumpable representation of the context
            back to the original context. This is used for json deserialization,
            which is being used in torch.export right now.
        flatten_with_keys_fn: An optional keyword argument to specify how to
            access each pytree leaf's keypath when flattening and tree-mapping.
            Like ``flatten_fn``, but in place of a List[leaf], it should return
            a List[(keypath, leaf)].
    """
    if to_str_fn is not None or maybe_from_str_fn is not None:
        warnings.warn(
            "`to_str_fn` and `maybe_from_str_fn` is deprecated. "
            "Please use `to_dumpable_context` and `from_dumpable_context` instead.",
            FutureWarning,
            stacklevel=2,
        )

    _private_register_pytree_node(
        cls,
        flatten_fn,
        unflatten_fn,
        serialized_type_name=serialized_type_name,
        to_dumpable_context=to_dumpable_context,
        from_dumpable_context=from_dumpable_context,
        flatten_with_keys_fn=flatten_with_keys_fn,
    )


def _private_register_pytree_node(
    cls: Type[Any],
    flatten_fn: FlattenFunc,
    unflatten_fn: UnflattenFunc,
    *,
    serialized_type_name: Optional[str] = None,
    to_dumpable_context: Optional[ToDumpableContextFn] = None,
    from_dumpable_context: Optional[FromDumpableContextFn] = None,
    flatten_with_keys_fn: Optional[FlattenWithKeysFunc] = None,
) -> None:
    """This is an internal function that is used to register a pytree node type
    for the Python pytree only. End-users should use :func:`register_pytree_node`
    instead.
    """
    with _NODE_REGISTRY_LOCK:
        if cls in SUPPORTED_NODES:
            # TODO: change this warning to an error after OSS/internal stabilize
            warnings.warn(
                f"{cls} is already registered as pytree node. "
                "Overwriting the previous registration.",
            )

        node_def = NodeDef(cls, flatten_fn, unflatten_fn, flatten_with_keys_fn)
        SUPPORTED_NODES[cls] = node_def

        if (to_dumpable_context is None) ^ (from_dumpable_context is None):
            raise ValueError(
                f"Both to_dumpable_context and from_dumpable_context for {cls} must "
                "be None or registered."
            )

        if serialized_type_name is None:
            serialized_type_name = NO_SERIALIZED_TYPE_NAME_FOUND

        serialize_node_def = _SerializeNodeDef(
            cls,
            serialized_type_name,
            to_dumpable_context,
            from_dumpable_context,
        )
        SUPPORTED_SERIALIZED_TYPES[cls] = serialize_node_def
        SERIALIZED_TYPE_TO_PYTHON_TYPE[serialized_type_name] = cls


@dataclasses.dataclass(frozen=True)
class SequenceKey(Generic[T]):
    idx: int

    def __str__(self) -> str:
        return f"[{self.idx!r}]"

    def get(self, sequence: Sequence[T]) -> T:
        return sequence[self.idx]


K = TypeVar("K", bound=Hashable)


@dataclasses.dataclass(frozen=True)
class MappingKey(Generic[K, T]):
    key: K

    def __str__(self) -> str:
        return f"[{self.key!r}]"

    def get(self, mapping: Mapping[K, T]) -> T:
        return mapping[self.key]


@dataclasses.dataclass(frozen=True)
class GetAttrKey:
    name: str

    def __str__(self) -> str:
        return f".{self.name}"

    def get(self, obj: Any) -> Any:
        return getattr(obj, self.name)


def _tuple_flatten(d: Tuple[Any, ...]) -> Tuple[List[Any], Context]:
    return list(d), None


def _tuple_flatten_with_keys(
    d: Tuple[Any, ...]
) -> Tuple[List[Tuple[KeyEntry, Any]], Context]:
    values, context = _tuple_flatten(d)
    return [(SequenceKey(i), v) for i, v in enumerate(values)], context


def _tuple_unflatten(values: Iterable[Any], context: Context) -> Tuple[Any, ...]:
    return tuple(values)


def _list_flatten(d: List[Any]) -> Tuple[List[Any], Context]:
    return d, None


def _list_flatten_with_keys(d: List[Any]) -> Tuple[List[Tuple[KeyEntry, Any]], Context]:
    values, context = _list_flatten(d)
    return [(SequenceKey(i), v) for i, v in enumerate(values)], context


def _list_unflatten(values: Iterable[Any], context: Context) -> List[Any]:
    return list(values)


def _dict_flatten(d: Dict[Any, Any]) -> Tuple[List[Any], Context]:
    return list(d.values()), list(d.keys())


def _dict_flatten_with_keys(
    d: Dict[Any, Any]
) -> Tuple[List[Tuple[KeyEntry, Any]], Context]:
    values, context = _dict_flatten(d)
    return [(MappingKey(k), v) for k, v in zip(context, values)], context


def _dict_unflatten(values: Iterable[Any], context: Context) -> Dict[Any, Any]:
    return dict(zip(context, values))


def _namedtuple_flatten(d: NamedTuple) -> Tuple[List[Any], Context]:
    return list(d), type(d)


def _namedtuple_flatten_with_keys(
    d: NamedTuple,
) -> Tuple[List[Tuple[KeyEntry, Any]], Context]:
    values, context = _namedtuple_flatten(d)
    return (
        [(GetAttrKey(field), v) for field, v in zip(context._fields, values)],
        context,
    )


def _namedtuple_unflatten(values: Iterable[Any], context: Context) -> NamedTuple:
    return cast(NamedTuple, context(*values))


def _namedtuple_serialize(context: Context) -> DumpableContext:
    if context not in SUPPORTED_SERIALIZED_TYPES:
        raise NotImplementedError(
            f"Can't serialize TreeSpec of namedtuple class {context} because we "
            "didn't register a serializated_type_name. Please register using "
            "`_register_namedtuple`."
        )

    serialize_node_def = SUPPORTED_SERIALIZED_TYPES[context]
    serialized_type_name = serialize_node_def.serialized_type_name

    if serialized_type_name == NO_SERIALIZED_TYPE_NAME_FOUND:
        raise NotImplementedError(
            f"Can't serialize TreeSpec of namedtuple class {context} because we "
            "couldn't find a serializated_type_name. Please register using "
            "`_register_namedtuple`."
        )
    return serialized_type_name


def _namedtuple_deserialize(dumpable_context: DumpableContext) -> Context:
    if dumpable_context not in SERIALIZED_TYPE_TO_PYTHON_TYPE:
        raise NotImplementedError(
            f"Can't deserialize TreeSpec of namedtuple class {dumpable_context} "
            "because we couldn't find a serializated name."
        )

    typ = SERIALIZED_TYPE_TO_PYTHON_TYPE[dumpable_context]
    return typ


def _ordereddict_flatten(d: GenericOrderedDict[Any, Any]) -> Tuple[List[Any], Context]:
    return list(d.values()), list(d.keys())


def _ordereddict_flatten_with_keys(
    d: GenericOrderedDict[Any, Any]
) -> Tuple[List[Tuple[KeyEntry, Any]], Context]:
    values, context = _ordereddict_flatten(d)
    return [(MappingKey(k), v) for k, v in zip(context, values)], context


def _ordereddict_unflatten(
    values: Iterable[Any],
    context: Context,
) -> GenericOrderedDict[Any, Any]:
    return OrderedDict((key, value) for key, value in zip(context, values))


_odict_flatten = _ordereddict_flatten
_odict_unflatten = _ordereddict_unflatten


def _defaultdict_flatten(d: DefaultDict[Any, Any]) -> Tuple[List[Any], Context]:
    values, dict_context = _dict_flatten(d)
    return values, [d.default_factory, dict_context]


def _defaultdict_flatten_with_keys(
    d: DefaultDict[Any, Any]
) -> Tuple[List[Tuple[KeyEntry, Any]], Context]:
    values, context = _defaultdict_flatten(d)
    _, dict_context = context
    return [(MappingKey(k), v) for k, v in zip(dict_context, values)], context


def _defaultdict_unflatten(
    values: Iterable[Any],
    context: Context,
) -> DefaultDict[Any, Any]:
    default_factory, dict_context = context
    return defaultdict(default_factory, _dict_unflatten(values, dict_context))


def _defaultdict_serialize(context: Context) -> DumpableContext:
    default_factory, dict_context = context
    json_defaultdict = {
        "default_factory_module": default_factory.__module__,
        "default_factory_name": default_factory.__qualname__,
        "dict_context": dict_context,
    }
    return json_defaultdict


def _defaultdict_deserialize(dumpable_context: DumpableContext) -> Context:
    assert isinstance(dumpable_context, dict)
    assert set(dumpable_context) == {
        "default_factory_module",
        "default_factory_name",
        "dict_context",
    }

    default_factory_module = dumpable_context["default_factory_module"]
    default_factory_name = dumpable_context["default_factory_name"]
    assert isinstance(default_factory_module, str)
    assert isinstance(default_factory_name, str)
    module = importlib.import_module(default_factory_module)
    default_factory = getattr(module, default_factory_name)

    dict_context = dumpable_context["dict_context"]
    return [default_factory, dict_context]


def _deque_flatten(d: Deque[Any]) -> Tuple[List[Any], Context]:
    return list(d), d.maxlen


def _deque_flatten_with_keys(
    d: Deque[Any],
) -> Tuple[List[Tuple[KeyEntry, Any]], Context]:
    values, context = _deque_flatten(d)
    return [(SequenceKey(i), v) for i, v in enumerate(values)], context


def _deque_unflatten(values: Iterable[Any], context: Context) -> Deque[Any]:
    return deque(values, maxlen=context)


_private_register_pytree_node(
    tuple,
    _tuple_flatten,
    _tuple_unflatten,
    serialized_type_name="builtins.tuple",
    flatten_with_keys_fn=_tuple_flatten_with_keys,
)
_private_register_pytree_node(
    list,
    _list_flatten,
    _list_unflatten,
    serialized_type_name="builtins.list",
    flatten_with_keys_fn=_list_flatten_with_keys,
)
_private_register_pytree_node(
    dict,
    _dict_flatten,
    _dict_unflatten,
    serialized_type_name="builtins.dict",
    flatten_with_keys_fn=_dict_flatten_with_keys,
)
_private_register_pytree_node(
    namedtuple,  # type: ignore[arg-type]
    _namedtuple_flatten,
    _namedtuple_unflatten,
    serialized_type_name="collections.namedtuple",
    to_dumpable_context=_namedtuple_serialize,
    from_dumpable_context=_namedtuple_deserialize,
    flatten_with_keys_fn=_namedtuple_flatten_with_keys,
)
_private_register_pytree_node(
    OrderedDict,
    _ordereddict_flatten,
    _ordereddict_unflatten,
    serialized_type_name="collections.OrderedDict",
    flatten_with_keys_fn=_ordereddict_flatten_with_keys,
)
_private_register_pytree_node(
    defaultdict,
    _defaultdict_flatten,
    _defaultdict_unflatten,
    serialized_type_name="collections.defaultdict",
    to_dumpable_context=_defaultdict_serialize,
    from_dumpable_context=_defaultdict_deserialize,
    flatten_with_keys_fn=_defaultdict_flatten_with_keys,
)
_private_register_pytree_node(
    deque,
    _deque_flatten,
    _deque_unflatten,
    serialized_type_name="collections.deque",
    flatten_with_keys_fn=_deque_flatten_with_keys,
)


STANDARD_DICT_TYPES: FrozenSet[type] = frozenset(
    {dict, OrderedDict, defaultdict},
)
BUILTIN_TYPES: FrozenSet[type] = frozenset(
    {tuple, list, dict, namedtuple, OrderedDict, defaultdict, deque},  # type: ignore[arg-type]
)


# h/t https://stackoverflow.com/questions/2166818/how-to-check-if-an-object-is-an-instance-of-a-namedtuple
def _is_namedtuple_instance(tree: Any) -> bool:
    typ = type(tree)
    bases = typ.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(typ, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(entry) == str for entry in fields)


def _get_node_type(tree: Any) -> Any:
    if _is_namedtuple_instance(tree):
        return namedtuple
    return type(tree)


# A leaf is defined as anything that is not a Node.
def _is_leaf(tree: PyTree, is_leaf: Optional[Callable[[PyTree], bool]] = None) -> bool:
    return (is_leaf is not None and is_leaf(tree)) or _get_node_type(
        tree
    ) not in SUPPORTED_NODES


# A TreeSpec represents the structure of a pytree. It holds:
# "type": the type of root Node of the pytree
# context: some context that is useful in unflattening the pytree
# children_specs: specs for each child of the root Node
# num_leaves: the number of leaves
@dataclasses.dataclass(init=True, frozen=True, eq=True, repr=False)
class TreeSpec:
    type: Any
    context: Context
    children_specs: List["TreeSpec"]

    num_nodes: int = dataclasses.field(init=False)
    num_leaves: int = dataclasses.field(init=False)
    num_children: int = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        num_nodes = sum((spec.num_nodes for spec in self.children_specs), start=1)
        num_leaves = sum(spec.num_leaves for spec in self.children_specs)
        num_children = len(self.children_specs)
        object.__setattr__(self, "num_nodes", num_nodes)
        object.__setattr__(self, "num_leaves", num_leaves)
        object.__setattr__(self, "num_children", num_children)

    def __repr__(self, indent: int = 0) -> str:
        repr_prefix: str = f"TreeSpec({self.type.__name__}, {self.context}, ["
        children_specs_str: str = ""
        if self.num_children > 0:
            indent += 2
            children_specs_str += self.children_specs[0].__repr__(indent)
            children_specs_str += "," if self.num_children > 1 else ""
            children_specs_str += ",".join(
                [
                    "\n" + " " * indent + child.__repr__(indent)
                    for child in self.children_specs[1:]
                ]
            )
        repr_suffix: str = f"{children_specs_str}])"
        return repr_prefix + repr_suffix

    def is_leaf(self) -> bool:
        return self.num_nodes == 1 and self.num_leaves == 1

    def _flatten_up_to_helper(self, tree: PyTree, subtrees: List[PyTree]) -> None:
        if self.is_leaf():
            subtrees.append(tree)
            return

        node_type = _get_node_type(tree)
        if self.type not in BUILTIN_TYPES:
            # Always require custom node types to match exactly
            if node_type != self.type:
                raise ValueError(
                    f"Type mismatch; "
                    f"expected {self.type!r}, but got {node_type!r}.",
                )
            flatten_fn = SUPPORTED_NODES[node_type].flatten_fn
            child_pytrees, context = flatten_fn(tree)
            if len(child_pytrees) != self.num_children:
                raise ValueError(
                    f"Node arity mismatch; "
                    f"expected {self.num_children}, but got {len(child_pytrees)}.",
                )
            if context != self.context:
                raise ValueError(
                    f"Node context mismatch for custom node type {self.type!r}.",
                )
        else:
            # For builtin dictionary types, we allow some flexibility
            # Otherwise, we require exact matches
            both_standard_dict = (
                self.type in STANDARD_DICT_TYPES and node_type in STANDARD_DICT_TYPES
            )
            if node_type != self.type and not both_standard_dict:
                raise ValueError(
                    f"Node type mismatch; "
                    f"expected {self.type!r}, but got {node_type!r}.",
                )
            if len(tree) != self.num_children:
                raise ValueError(
                    f"Node arity mismatch; "
                    f"expected {self.num_children}, but got {len(tree)}.",
                )

            if both_standard_dict:  # dictionary types are compatible with each other
                dict_context = (
                    self.context
                    if self.type is not defaultdict
                    # ignore mismatch of `default_factory` for defaultdict
                    else self.context[1]
                )
                expected_keys = dict_context
                got_key_set = set(tree)
                expected_key_set = set(expected_keys)
                if got_key_set != expected_key_set:
                    missing_keys = expected_key_set.difference(got_key_set)
                    extra_keys = got_key_set.difference(expected_key_set)
                    message = ""
                    if missing_keys:
                        message += f"; missing key(s): {missing_keys}"
                    if extra_keys:
                        message += f"; extra key(s): {extra_keys}"
                    raise ValueError(f"Node keys mismatch{message}.")
                child_pytrees = [tree[key] for key in expected_keys]
            else:
                flatten_fn = SUPPORTED_NODES[node_type].flatten_fn
                child_pytrees, context = flatten_fn(tree)
                if (
                    context != self.context
                    and self.type is not deque  # ignore mismatch of `maxlen` for deque
                ):
                    raise ValueError(
                        f"Node context mismatch for node type {self.type!r}; "
                        f"expected {self.context!r}, but got {context!r}.",  # namedtuple type mismatch
                    )

        for child_pytree, child_spec in zip(child_pytrees, self.children_specs):
            child_spec._flatten_up_to_helper(child_pytree, subtrees)

    def flatten_up_to(self, tree: PyTree) -> List[PyTree]:
        subtrees: List[PyTree] = []
        self._flatten_up_to_helper(tree, subtrees)
        return subtrees

    def unflatten(self, leaves: Iterable[Any]) -> PyTree:
        if not isinstance(leaves, (list, tuple)):
            leaves = list(leaves)
        if len(leaves) != self.num_leaves:
            raise ValueError(
                f"treespec.unflatten(leaves): `leaves` has length {len(leaves)} "
                f"but the spec refers to a pytree that holds {self.num_leaves} "
                f"items ({self}).",
            )
        if self.is_leaf():
            return leaves[0]

        unflatten_fn = SUPPORTED_NODES[self.type].unflatten_fn

        # Recursively unflatten the children
        start = 0
        end = 0
        child_pytrees = []
        for child_spec in self.children_specs:
            end += child_spec.num_leaves
            child_pytrees.append(child_spec.unflatten(leaves[start:end]))
            start = end

        return unflatten_fn(child_pytrees, self.context)


class LeafSpec(TreeSpec):
    def __init__(self) -> None:
        super().__init__(None, None, [])

    def __post_init__(self) -> None:
        object.__setattr__(self, "num_nodes", 1)
        object.__setattr__(self, "num_leaves", 1)
        object.__setattr__(self, "num_children", 0)

    def __repr__(self, indent: int = 0) -> str:
        return "*"


# All leaves are equivalent, so represent with a single object to save on
# object construction time
_LEAF_SPEC = LeafSpec()


def _tree_flatten_helper(
    tree: PyTree,
    leaves: List[Any],
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> TreeSpec:
    if _is_leaf(tree, is_leaf=is_leaf):
        leaves.append(tree)
        return _LEAF_SPEC

    node_type = _get_node_type(tree)
    flatten_fn = SUPPORTED_NODES[node_type].flatten_fn
    child_pytrees, context = flatten_fn(tree)

    # Recursively flatten the children
    children_specs = [
        _tree_flatten_helper(child, leaves, is_leaf=is_leaf) for child in child_pytrees
    ]

    return TreeSpec(node_type, context, children_specs)


def tree_flatten(
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> Tuple[List[Any], TreeSpec]:
    """Flattens a pytree into a list of values and a TreeSpec that can be used
    to reconstruct the pytree.
    """
    leaves: List[Any] = []
    spec = _tree_flatten_helper(tree, leaves, is_leaf=is_leaf)
    return leaves, spec


def tree_unflatten(leaves: Iterable[Any], treespec: TreeSpec) -> PyTree:
    """Given a list of values and a TreeSpec, builds a pytree.
    This is the inverse operation of `tree_flatten`.
    """
    if not isinstance(treespec, TreeSpec):
        raise TypeError(
            f"tree_unflatten(leaves, treespec): Expected `treespec` to be "
            f"instance of TreeSpec but got item of type {type(treespec)}.",
        )
    return treespec.unflatten(leaves)


def tree_iter(
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> Iterable[Any]:
    """Get an iterator over the leaves of a pytree."""
    if _is_leaf(tree, is_leaf=is_leaf):
        yield tree
    else:
        node_type = _get_node_type(tree)
        flatten_fn = SUPPORTED_NODES[node_type].flatten_fn
        child_pytrees, _ = flatten_fn(tree)

        # Recursively flatten the children
        for child in child_pytrees:
            yield from tree_iter(child, is_leaf=is_leaf)


def tree_leaves(
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> List[Any]:
    """Get a list of leaves of a pytree."""
    return list(tree_iter(tree, is_leaf=is_leaf))


def tree_structure(
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> TreeSpec:
    """Get the TreeSpec for a pytree."""
    return tree_flatten(tree, is_leaf=is_leaf)[1]


def tree_map(
    func: Callable[..., Any],
    tree: PyTree,
    *rests: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    """Map a multi-input function over pytree args to produce a new pytree.

    See also :func:`tree_map_`.

    >>> tree_map(lambda x: x + 1, {'x': 7, 'y': (42, 64)})
    {'x': 8, 'y': (43, 65)}
    >>> tree_map(lambda x: x is None, {'x': 7, 'y': (42, 64), 'z': None})
    {'x': False, 'y': (False, False), 'z': True}

    If multiple inputs are given, the structure of the tree is taken from the first input;
    subsequent inputs need only have ``tree`` as a prefix:

    >>> tree_map(lambda x, y: [x] + y, [5, 6], [[7, 9], [1, 2]])
    [[5, 7, 9], [6, 1, 2]]

    Args:
        func (callable): A function that takes ``1 + len(rests)`` arguments, to be applied at the
            corresponding leaves of the pytrees.
        tree (pytree): A pytree to be mapped over, with each leaf providing the first positional
            argument to function ``func``.
        rests (tuple of pytree): A tuple of pytrees, each of which has the same structure as
            ``tree`` or has ``tree`` as a prefix.
        is_leaf (callable, optional): An extra leaf predicate function that will be called at each
            flattening step. The function should have a single argument with signature
            ``is_leaf(node) -> bool``. If it returns :data:`True`, the whole subtree being treated
            as a leaf. Otherwise, the default pytree registry will be used to determine a node is a
            leaf or not. If the function is not specified, the default pytree registry will be used.

    Returns:
        A new pytree with the same structure as ``tree`` but with the value at each leaf given by
        ``func(x, *xs)`` where ``x`` is the value at the corresponding leaf in ``tree`` and ``xs``
        is the tuple of values at corresponding nodes in ``rests``.
    """
    leaves, treespec = tree_flatten(tree, is_leaf=is_leaf)
    flat_args = [leaves] + [treespec.flatten_up_to(r) for r in rests]
    return treespec.unflatten(map(func, *flat_args))


def tree_map_(
    func: Callable[..., Any],
    tree: PyTree,
    *rests: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    """Like :func:`tree_map`, but do an inplace call on each leaf and return the original tree.

    See also :func:`tree_map`.

    Args:
        func (callable): A function that takes ``1 + len(rests)`` arguments, to be applied at the
            corresponding leaves of the pytrees.
        tree (pytree): A pytree to be mapped over, with each leaf providing the first positional
            argument to function ``func``.
        rests (tuple of pytree): A tuple of pytrees, each of which has the same structure as
            ``tree`` or has ``tree`` as a prefix.
        is_leaf (callable, optional): An extra leaf predicate function that will be called at each
            flattening step. The function should have a single argument with signature
            ``is_leaf(node) -> bool``. If it returns :data:`True`, the whole subtree being treated
            as a leaf. Otherwise, the default pytree registry will be used to determine a node is a
            leaf or not. If the function is not specified, the default pytree registry will be used.

    Returns:
        The original ``tree`` with the value at each leaf is given by the side-effect of function
        ``func(x, *xs)`` (not the return value) where ``x`` is the value at the corresponding leaf
        in ``tree`` and ``xs`` is the tuple of values at values at corresponding nodes in ``rests``.
    """
    leaves, treespec = tree_flatten(tree, is_leaf=is_leaf)
    flat_args = [leaves] + [treespec.flatten_up_to(r) for r in rests]
    tuple(map(func, *flat_args))  # consume and exhaust the iterable
    return tree


Type2 = Tuple[Type[T], Type[S]]
Type3 = Tuple[Type[T], Type[S], Type[U]]
if sys.version_info >= (3, 10):
    TypeAny = Union[Type[Any], Tuple[Type[Any], ...], types.UnionType]
else:
    TypeAny = Union[Type[Any], Tuple[Type[Any], ...]]

Fn2 = Callable[[Union[T, S]], R]
Fn3 = Callable[[Union[T, S, U]], R]
Fn = Callable[[T], R]
FnAny = Callable[[Any], R]

MapOnlyFn = Callable[[T], Callable[[Any], Any]]


# These specializations help with type inference on the lambda passed to this
# function
@overload
def map_only(__type_or_types_or_pred: Type2[T, S]) -> MapOnlyFn[Fn2[T, S, Any]]:
    ...


@overload
def map_only(__type_or_types_or_pred: Type3[T, S, U]) -> MapOnlyFn[Fn3[T, S, U, Any]]:
    ...


@overload
def map_only(__type_or_types_or_pred: Type[T]) -> MapOnlyFn[Fn[T, Any]]:
    ...


# This specialization is needed for the implementations below that call
@overload
def map_only(__type_or_types_or_pred: TypeAny) -> MapOnlyFn[FnAny[Any]]:
    ...


@overload
def map_only(__type_or_types_or_pred: Callable[[Any], bool]) -> MapOnlyFn[FnAny[Any]]:
    ...


def map_only(
    __type_or_types_or_pred: Union[TypeAny, Callable[[Any], bool]]
) -> MapOnlyFn[FnAny[Any]]:
    """
    Suppose you are writing a tree_map over tensors, leaving everything
    else unchanged.  Ordinarily you would have to write:

        def go(t):
            if isinstance(t, Tensor):
                return ...
            else:
                return t

    With this function, you only need to write:

        @map_only(Tensor)
        def go(t):
            return ...

    You can also directly use 'tree_map_only'
    """
    if isinstance(__type_or_types_or_pred, (type, tuple)) or (
        sys.version_info >= (3, 10)
        and isinstance(__type_or_types_or_pred, types.UnionType)
    ):

        def pred(x: Any) -> bool:
            return isinstance(x, __type_or_types_or_pred)  # type: ignore[arg-type]

    elif callable(__type_or_types_or_pred):
        pred = __type_or_types_or_pred  # type: ignore[assignment]
    else:
        raise TypeError("Argument must be a type, a tuple of types, or a callable.")

    def wrapper(func: Callable[[T], Any]) -> Callable[[Any], Any]:
        @functools.wraps(func)
        def wrapped(x: T) -> Any:
            if pred(x):
                return func(x)
            return x

        return wrapped

    return wrapper


@overload
def tree_map_only(
    __type_or_types_or_pred: Type[T],
    func: Fn[T, Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    ...


@overload
def tree_map_only(
    __type_or_types_or_pred: Type2[T, S],
    func: Fn2[T, S, Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    ...


@overload
def tree_map_only(
    __type_or_types_or_pred: Type3[T, S, U],
    func: Fn3[T, S, U, Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    ...


@overload
def tree_map_only(
    __type_or_types_or_pred: Callable[[Any], bool],
    func: FnAny[Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    ...


def tree_map_only(
    __type_or_types_or_pred: Union[TypeAny, Callable[[Any], bool]],
    func: FnAny[Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    return tree_map(map_only(__type_or_types_or_pred)(func), tree, is_leaf=is_leaf)


@overload
def tree_map_only_(
    __type_or_types_or_pred: Type[T],
    func: Fn[T, Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    ...


@overload
def tree_map_only_(
    __type_or_types_or_pred: Type2[T, S],
    func: Fn2[T, S, Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    ...


@overload
def tree_map_only_(
    __type_or_types_or_pred: Type3[T, S, U],
    func: Fn3[T, S, U, Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    ...


@overload
def tree_map_only_(
    __type_or_types_or_pred: Callable[[Any], bool],
    func: FnAny[Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    ...


def tree_map_only_(
    __type_or_types_or_pred: Union[TypeAny, Callable[[Any], bool]],
    func: FnAny[Any],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    return tree_map_(map_only(__type_or_types_or_pred)(func), tree, is_leaf=is_leaf)


def tree_all(
    pred: Callable[[Any], bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    flat_args = tree_iter(tree, is_leaf=is_leaf)
    return all(map(pred, flat_args))


def tree_any(
    pred: Callable[[Any], bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    flat_args = tree_iter(tree, is_leaf=is_leaf)
    return any(map(pred, flat_args))


@overload
def tree_all_only(
    __type_or_types: Type[T],
    pred: Fn[T, bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    ...


@overload
def tree_all_only(
    __type_or_types: Type2[T, S],
    pred: Fn2[T, S, bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    ...


@overload
def tree_all_only(
    __type_or_types: Type3[T, S, U],
    pred: Fn3[T, S, U, bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    ...


def tree_all_only(
    __type_or_types: TypeAny,
    pred: FnAny[bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    flat_args = tree_iter(tree, is_leaf=is_leaf)
    return all(pred(x) for x in flat_args if isinstance(x, __type_or_types))


@overload
def tree_any_only(
    __type_or_types: Type[T],
    pred: Fn[T, bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    ...


@overload
def tree_any_only(
    __type_or_types: Type2[T, S],
    pred: Fn2[T, S, bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    ...


@overload
def tree_any_only(
    __type_or_types: Type3[T, S, U],
    pred: Fn3[T, S, U, bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    ...


def tree_any_only(
    __type_or_types: TypeAny,
    pred: FnAny[bool],
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> bool:
    flat_args = tree_iter(tree, is_leaf=is_leaf)
    return any(pred(x) for x in flat_args if isinstance(x, __type_or_types))


# Broadcasts a pytree to the provided TreeSpec and returns the flattened
# values. If this is not possible, then this function returns None.
#
# For example, given pytree=0 and spec=TreeSpec(list, None, [LeafSpec(), LeafSpec()]),
# would return [0, 0]. This is useful for part of the vmap implementation:
# a user can pass in vmap(fn, in_dims)(*inputs). `in_dims` should be
# broadcastable to the tree structure of `inputs` and we use
# _broadcast_to_and_flatten to check this.
def _broadcast_to_and_flatten(
    tree: PyTree,
    treespec: TreeSpec,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> Optional[List[Any]]:
    assert isinstance(treespec, TreeSpec)

    if _is_leaf(tree, is_leaf=is_leaf):
        return [tree] * treespec.num_leaves
    if treespec.is_leaf():
        return None
    node_type = _get_node_type(tree)
    if node_type != treespec.type:
        return None

    flatten_fn = SUPPORTED_NODES[node_type].flatten_fn
    child_pytrees, ctx = flatten_fn(tree)

    # Check if the Node is different from the spec
    if len(child_pytrees) != treespec.num_children or ctx != treespec.context:
        return None

    # Recursively flatten the children
    result: List[Any] = []
    for child, child_spec in zip(child_pytrees, treespec.children_specs):
        flat = _broadcast_to_and_flatten(child, child_spec, is_leaf=is_leaf)
        if flat is not None:
            result += flat
        else:
            return None

    return result


@dataclasses.dataclass
class _TreeSpecSchema:
    """
    _TreeSpecSchema is the schema used to serialize the TreeSpec
    It contains the following fields:
    - type: A string name of the type. null for the case of a LeafSpec.
    - context: Any format which is json dumpable
    - children_spec: A list of children serialized specs.
    """

    type: Optional[str]
    context: DumpableContext
    children_spec: List["_TreeSpecSchema"]


class _ProtocolFn(NamedTuple):
    treespec_to_json: Callable[[TreeSpec], DumpableContext]
    json_to_treespec: Callable[[DumpableContext], TreeSpec]


_SUPPORTED_PROTOCOLS: Dict[int, _ProtocolFn] = {}


def _treespec_to_json(treespec: TreeSpec) -> _TreeSpecSchema:
    if treespec.is_leaf():
        return _TreeSpecSchema(None, None, [])

    if treespec.type not in SUPPORTED_SERIALIZED_TYPES:
        raise NotImplementedError(
            f"Serializing {treespec.type} in pytree is not registered.",
        )

    serialize_node_def = SUPPORTED_SERIALIZED_TYPES[treespec.type]

    serialized_type_name = serialize_node_def.serialized_type_name

    if serialized_type_name == NO_SERIALIZED_TYPE_NAME_FOUND:
        raise NotImplementedError(
            f"No registered serialization name for {treespec.type} found. "
            "Please update your _register_pytree_node call with a `serialized_type_name` kwarg."
        )

    if serialize_node_def.to_dumpable_context is None:
        try:
            serialized_context = json.dumps(treespec.context)
        except TypeError as e:
            raise TypeError(
                "Unable to serialize context. "
                "Please make the context json dump-able, or register a "
                "custom serializer using _register_pytree_node."
            ) from e
    else:
        serialized_context = serialize_node_def.to_dumpable_context(treespec.context)

    child_schemas = [_treespec_to_json(child) for child in treespec.children_specs]

    return _TreeSpecSchema(serialized_type_name, serialized_context, child_schemas)


def _json_to_treespec(json_schema: DumpableContext) -> TreeSpec:
    if (
        json_schema["type"] is None
        and json_schema["context"] is None
        and len(json_schema["children_spec"]) == 0
    ):
        return _LEAF_SPEC

    if json_schema["type"] not in SERIALIZED_TYPE_TO_PYTHON_TYPE:
        raise NotImplementedError(
            f'Deserializing {json_schema["type"]} in pytree is not registered.',
        )

    typ = SERIALIZED_TYPE_TO_PYTHON_TYPE[json_schema["type"]]
    serialize_node_def = SUPPORTED_SERIALIZED_TYPES[typ]

    if serialize_node_def.from_dumpable_context is None:
        try:
            context = json.loads(json_schema["context"])
        except TypeError as ex:
            raise TypeError(
                "Unable to deserialize context. "
                "Please make the context json load-able, or register a "
                "custom serializer using _register_pytree_node.",
            ) from ex
    else:
        context = serialize_node_def.from_dumpable_context(json_schema["context"])

    children_specs = []
    for child_string in json_schema["children_spec"]:
        children_specs.append(_json_to_treespec(child_string))

    return TreeSpec(typ, context, children_specs)


_SUPPORTED_PROTOCOLS[1] = _ProtocolFn(_treespec_to_json, _json_to_treespec)


def treespec_dumps(treespec: TreeSpec, protocol: Optional[int] = None) -> str:
    if not isinstance(treespec, TreeSpec):
        raise TypeError(
            f"treespec_dumps(treespec, protocol): Expected `treespec` to be instance of "
            f"TreeSpec but got item of type {type(treespec)}.",
        )

    if protocol is None:
        protocol = DEFAULT_TREESPEC_SERIALIZATION_PROTOCOL

    if protocol in _SUPPORTED_PROTOCOLS:
        json_spec = _SUPPORTED_PROTOCOLS[protocol].treespec_to_json(treespec)
    else:
        raise ValueError(
            f"Unknown protocol {protocol}. "
            f"Available protocols: {list(_SUPPORTED_PROTOCOLS.keys())}",
        )

    str_spec = json.dumps((protocol, dataclasses.asdict(json_spec)))
    return str_spec


def treespec_loads(serialized: str) -> TreeSpec:
    protocol, json_schema = json.loads(serialized)

    if protocol in _SUPPORTED_PROTOCOLS:
        return _SUPPORTED_PROTOCOLS[protocol].json_to_treespec(json_schema)
    raise ValueError(
        f"Unknown protocol {protocol}. "
        f"Available protocols: {list(_SUPPORTED_PROTOCOLS.keys())}",
    )


class _DummyLeaf:
    def __repr__(self) -> str:
        return "*"


def treespec_pprint(treespec: TreeSpec) -> str:
    dummy_tree = tree_unflatten(
        [_DummyLeaf() for _ in range(treespec.num_leaves)],
        treespec,
    )
    return repr(dummy_tree)


# TODO(angelayi): remove this function after OSS/internal stabilize
@deprecated(
    "`pytree_to_str` is deprecated. Please use `treespec_dumps` instead.",
    category=FutureWarning,
)
def pytree_to_str(treespec: TreeSpec) -> str:
    return treespec_dumps(treespec)


# TODO(angelayi): remove this function after OSS/internal stabilize
@deprecated(
    "`str_to_pytree` is deprecated. Please use `treespec_loads` instead.",
    category=FutureWarning,
)
def str_to_pytree(json: str) -> TreeSpec:
    return treespec_loads(json)


def arg_tree_leaves(*args: PyTree, **kwargs: PyTree) -> List[Any]:
    """Get a flat list of arguments to this function

    A slightly faster version of tree_leaves((args, kwargs))
    """
    leaves: List[Any] = []
    for a in args:
        leaves.extend(tree_iter(a))
    for a in kwargs.values():
        leaves.extend(tree_iter(a))
    return leaves


def tree_flatten_with_path(
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> Tuple[List[Tuple[KeyPath, Any]], TreeSpec]:
    """Flattens a pytree like :func:`tree_flatten`, but also returns each leaf's key path.

    Args:
        tree: a pytree to flatten. If it contains a custom type, that type must be
            registered with an appropriate `tree_flatten_with_path_fn` when registered
            with :func:`register_pytree_node`.
        is_leaf: An extra leaf predicate function that will be called at each
            flattening step. The function should have a single argument with signature
            ``is_leaf(node) -> bool``. If it returns :data:`True`, the whole subtree being treated
            as a leaf. Otherwise, the default pytree registry will be used to determine a node is a
            leaf or not. If the function is not specified, the default pytree registry will be used.
    Returns:
        A tuple where the first element is a list of (key path, leaf) pairs, and the
        second element is a :class:`TreeSpec` representing the structure of the flattened
        tree.
    """
    _, treespec = tree_flatten(tree, is_leaf)
    return list(_generate_key_paths((), tree, is_leaf)), treespec


def tree_leaves_with_path(
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> List[Tuple[KeyPath, Any]]:
    """Gets the leaves of a pytree like ``tree_leaves`` and returns each leaf's key path.

    Args:
        tree: a pytree. If it contains a custom type, that type must be
            registered with an appropriate `tree_flatten_with_path_fn` when registered
            with :func:`register_pytree_node`.
        is_leaf: An extra leaf predicate function that will be called at each
            flattening step. The function should have a single argument with signature
            ``is_leaf(node) -> bool``. If it returns :data:`True`, the whole subtree being treated
            as a leaf. Otherwise, the default pytree registry will be used to determine a node is a
            leaf or not. If the function is not specified, the default pytree registry will be used.
    Returns:
        A list of (key path, leaf) pairs.
    """
    return list(_generate_key_paths((), tree, is_leaf))


def _generate_key_paths(
    key_path: KeyPath,
    tree: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> Iterable[Tuple[KeyPath, Any]]:
    if is_leaf and is_leaf(tree):
        yield key_path, tree
        return

    node_type = _get_node_type(tree)
    handler = SUPPORTED_NODES.get(node_type)
    if not handler:
        # This is a leaf
        yield key_path, tree
        return

    flatten_with_keys = handler.flatten_with_keys_fn
    if flatten_with_keys:
        key_children, _ = flatten_with_keys(tree)
        for k, c in key_children:
            yield from _generate_key_paths((*key_path, k), c, is_leaf)
    else:
        # We registered this pytree but didn't add a flatten_with_keys_fn, complain.
        raise ValueError(
            f"Did not find a flatten_with_keys_fn for type: {node_type}. "
            "Please pass a flatten_with_keys_fn argument to register_pytree_node."
        )


def tree_map_with_path(
    func: Callable[..., Any],
    tree: PyTree,
    *rests: PyTree,
    is_leaf: Optional[Callable[[PyTree], bool]] = None,
) -> PyTree:
    """Like :func:`tree_map`, but the provided callable takes an additional key path argument.

    Args:
        func: A function that takes ``2 + len(rests)`` arguments, to be applied at the
            corresponding leaves of the pytrees. The first positional argument
            to ``func`` is the key path of the leaf in question. The second
            positional argument is the value of the leaf.
        tree: A pytree to be mapped over, with each leaf providing the first positional
            argument to function ``func``.
        rests: A tuple of pytrees, each of which has the same structure as
            ``tree`` or has ``tree`` as a prefix.
        is_leaf: An extra leaf predicate function that will be called at each
            flattening step. The function should have a single argument with signature
            ``is_leaf(node) -> bool``. If it returns :data:`True`, the whole subtree being treated
            as a leaf. Otherwise, the default pytree registry will be used to determine a node is a
            leaf or not. If the function is not specified, the default pytree registry will be used.

    Returns
        A new pytree with the same structure as ``tree`` but with the value at each leaf given by
        ``func(keypath, x, *xs)`` where ``keypath`` is the key path at the
        corresponding leaf in ``tree``, ``x`` is the value at that leaf, and
        ``xs`` is the tuple of values at corresponding nodes in ``rests``.
    """
    keypath_leaves, treespec = tree_flatten_with_path(tree, is_leaf)
    keypath_leaves = list(zip(*keypath_leaves))
    all_keypath_leaves = keypath_leaves + [treespec.flatten_up_to(r) for r in rests]
    return treespec.unflatten(func(*xs) for xs in zip(*all_keypath_leaves))


def keystr(kp: KeyPath) -> str:
    """Given a key path, return a pretty-printed representation."""
    return "".join([str(k) for k in kp])


def key_get(obj: Any, kp: KeyPath) -> Any:
    """Given an object and a key path, return the value at the key path."""
    for k in kp:
        obj = k.get(obj)
    return obj



from abc import ABC, abstractmethod
from typing import List

class BaseRecommender(ABC):
    @abstractmethod
    def fit(self, data):
        """
        Train the recommender system using the provided data.
        
        Parameters:
        - data: The data used for training the model.
        """
        pass

    @abstractmethod
    def recommend(self, item_indices: List[int], top_n: int = 10) -> List[int]:
        """
        Generate top-N item recommendations based on content features.
        
        Parameters:
        - item_indices (List[int]): List of item indices to base recommendations on.
        - top_n (int): The number of recommendations to generate.
        
        Returns:
        - List[int]: List of recommended item indices.
        """
        pass



# corerec/engines/unionizedFilterEngine/mf_base/matrix_factorization_base.py
import numpy as np
from scipy.sparse import csr_matrix
from typing import List, Optional
from corerec.engines.unionizedFilterEngine.base_recommender import BaseRecommender
from concurrent.futures import ThreadPoolExecutor

class MatrixFactorizationBase(BaseRecommender):
    """
    MatrixFactorizationBase: A Scalable and Efficient Recommender System
    Author: Vishesh Yadav
    
    Overview:
    ----------
    The `MatrixFactorizationBase` class implements a matrix factorization-based
    recommendation system using stochastic gradient descent (SGD) for optimization.
    It is designed to be efficient and scalable, suitable for large-scale datasets.

    Key Features:
    --------------
    - **Vectorized Operations**: Utilizes NumPy's vectorized operations to perform
      computations efficiently, reducing the need for explicit loops.
    
    - **Bias Terms**: Incorporates user and item bias terms to capture inherent
      biases, improving prediction accuracy.
    
    - **Sparse Matrix Handling**: Leverages SciPy's sparse matrix capabilities to
      efficiently manage large, sparse interaction matrices, minimizing memory usage.
    
    - **Stochastic Gradient Descent (SGD)**: Employs SGD for parameter updates,
      enabling faster convergence and scalability with large datasets.
    
    - **Parallelized Factor Updates**: Utilizes multi-threading to update user and
      item factors in parallel, taking advantage of multi-core processors.
    
    - **Early Stopping**: Includes an early stopping mechanism based on validation
      loss to prevent overfitting and reduce unnecessary epochs.
    
    - **Regularization**: Provides separate regularization parameters for user and
      item factors, allowing fine control over model complexity.
    
    - **Xavier Initialization**: Uses Xavier initialization for factor initialization,
      ensuring better convergence properties.

    Parameters:
    -------------
    - `num_factors` (int): Number of latent factors for users and items.
    - `learning_rate` (float): Learning rate for SGD updates.
    - `reg_user` (float): Regularization parameter for user factors and biases.
    - `reg_item` (float): Regularization parameter for item factors and biases.
    - `epochs` (int): Number of training epochs.
    - `early_stopping_rounds` (Optional[int]): Number of epochs with no improvement
      on validation loss to trigger early stopping.
    - `n_threads` (int): Number of threads for parallel processing.

    Methods:
    ---------
    - `initialize_factors(num_users, num_items)`: Initializes user and item factors
      and biases using Xavier initialization.
    
    - `compute_loss(interaction_matrix)`: Computes the loss (MSE + regularization)
      for the given interaction matrix.
    
    - `fit(interaction_matrix, validation_matrix)`: Trains the model using the
      interaction matrix, with optional validation for early stopping.
    
    - `_sgd_step(interaction_matrix)`: Performs a single SGD step to update user
      and item factors and biases.

    Usage:
    -------
    This class is intended for use in recommendation systems where scalability and
    efficiency are critical. It is particularly well-suited for large datasets with
    sparse interactions, such as user-item ratings in collaborative filtering tasks.
    """
    def __init__(self, num_factors: int = 20, learning_rate: float = 0.01, 
                 reg_user: float = 0.02, reg_item: float = 0.02, 
                 epochs: int = 20, early_stopping_rounds: Optional[int] = None, 
                 n_threads: int = 4):
        self.num_factors = num_factors
        self.learning_rate = learning_rate
        self.reg_user = reg_user
        self.reg_item = reg_item
        self.epochs = epochs
        self.early_stopping_rounds = early_stopping_rounds
        self.n_threads = n_threads
        self.user_factors = None
        self.item_factors = None
        self.user_bias = None
        self.item_bias = None
        self.global_bias = 0.0

    def initialize_factors(self, num_users: int, num_items: int):
        # Xavier Initialization for better convergence
        limit = np.sqrt(6 / (num_users + self.num_factors))
        self.user_factors = np.random.uniform(-limit, limit, (num_users, self.num_factors))
        self.item_factors = np.random.uniform(-limit, limit, (num_items, self.num_factors))
        self.user_bias = np.zeros(num_users)
        self.item_bias = np.zeros(num_items)
        self.global_bias = 0.0

    def compute_loss(self, interaction_matrix: csr_matrix) -> float:
        # Vectorized computation of predictions
        predictions = self.user_factors.dot(self.item_factors.T) + self.user_bias[:, np.newaxis] + self.item_bias[np.newaxis, :] + self.global_bias
        errors = interaction_matrix.toarray() - predictions
        mse = np.sum(errors ** 2) / interaction_matrix.nnz
        reg = (self.reg_user * np.sum(self.user_factors ** 2) +
               self.reg_item * np.sum(self.item_factors ** 2) +
               self.reg_user * np.sum(self.user_bias ** 2) +
               self.reg_item * np.sum(self.item_bias ** 2))
        return mse + reg

    def fit(self, interaction_matrix: csr_matrix, validation_matrix: Optional[csr_matrix] = None):
        num_users, num_items = interaction_matrix.shape
        self.initialize_factors(num_users, num_items)
        self.global_bias = interaction_matrix.data.mean()
        
        best_loss = float('inf')
        no_improve_epochs = 0

        for epoch in range(self.epochs):
            self._sgd_step(interaction_matrix)
            loss = self.compute_loss(interaction_matrix)
            print(f"Epoch {epoch + 1}/{self.epochs}, Loss: {loss:.4f}")

            if validation_matrix is not None:
                val_loss = self.compute_loss(validation_matrix)
                print(f"Validation Loss: {val_loss:.4f}")
                if val_loss < best_loss:
                    best_loss = val_loss
                    no_improve_epochs = 0
                else:
                    no_improve_epochs += 1
                    if self.early_stopping_rounds and no_improve_epochs >= self.early_stopping_rounds:
                        print("Early stopping triggered.")
                        break

    def _sgd_step(self, interaction_matrix: csr_matrix):
        """
        Perform a single SGD step to update user and item factors and biases.
        This implementation uses multi-threading to parallelize updates.

        Parameters:
        - interaction_matrix (csr_matrix): User-item interaction matrix.
        """
        def update_user(u):
            user_interactions = interaction_matrix[u].indices
            for i in user_interactions:
                prediction = np.dot(self.user_factors[u], self.item_factors[i]) + self.user_bias[u] + self.item_bias[i] + self.global_bias
                error = interaction_matrix[u, i] - prediction
                # Update biases
                self.user_bias[u] += self.learning_rate * (error - self.reg_user * self.user_bias[u])
                self.item_bias[i] += self.learning_rate * (error - self.reg_item * self.item_bias[i])
                # Update latent factors
                self.user_factors[u] += self.learning_rate * (error * self.item_factors[i] - self.reg_user * self.user_factors[u])
                self.item_factors[i] += self.learning_rate * (error * self.user_factors[u] - self.reg_item * self.item_factors[i])

        with ThreadPoolExecutor(max_workers=self.n_threads) as executor:
            executor.map(update_user, range(interaction_matrix.shape[0]))


# corerec/engines/unionizedFilterEngine/nn_base/neural_cf_base.py

from abc import ABC
from typing import List
from scipy.sparse import csr_matrix
from ..base_recommender import BaseRecommender

class NeuralCollaborativeFilteringBase(BaseRecommender, ABC):
    """
    Neural Collaborative Filtering (NCF) Base Implementation.

    A generalized framework for neural network-based collaborative filtering that can learn
    arbitrary function from data by leveraging neural networks' capacity of non-linear 
    transformation and deep representation learning.

    Attributes:
        embedding_size (int): Dimension of embedding vectors
        learning_rate (float): Learning rate for optimization
        epochs (int): Number of training epochs
        layers (List[int]): Architecture of neural layers
        batch_size (int): Size of mini-batches
        optimizer (str): Optimization algorithm

    Features:
        - Flexible neural architectures
        - Multi-layer perceptron for feature learning
        - Customizable embedding dimensions
        - Mini-batch training support
        - Multiple loss function options
        - Regularization techniques

    Training Process:
        1. Embedding Layer: Maps sparse features to dense vectors
        2. Feature Learning: Multiple neural layers for interaction
        3. Prediction Layer: Final layer for recommendation scores
        4. Optimization: Minimizes loss function through backpropagation

    Mathematical Formulation:
        For a user-item pair (u,i):
        1. User embedding: p_u = P * v_u
        2. Item embedding: q_i = Q * v_i
        3. Neural CF layers: φ_out = φ_L(...φ_2(φ_1([p_u, q_i]))...)
        4. Prediction: ŷ_ui = σ(h^T * φ_out)

    References:
        He, X., et al. "Neural Collaborative Filtering." WWW 2017.
    """

    def __init__(self, embedding_size: int = 50, learning_rate: float = 0.001, epochs: int = 10):
        """
        Initialize the Neural Collaborative Filtering model.

        Args:
            embedding_size (int): Dimension of the embedding vectors
            learning_rate (float): Learning rate for optimization
            epochs (int): Number of training epochs

        Note:
            The embedding size should be chosen based on the dataset size
            and complexity of the recommendation task.
        """
        self.embedding_size = embedding_size
        self.learning_rate = learning_rate
        self.epochs = epochs
        # Add additional initialization as needed

    def preprocess_data(self, interaction_matrix: csr_matrix, user_ids: List[int], item_ids: List[int]):
        # Implement any preprocessing steps specific to neural CF
        pass


from typing import Dict, Any
from .base_recommender import BaseRecommender
from .tfidf_recommender import TFIDFRecommender
# Import other content-based recommender classes here as they are implemented

class ContentFilterFactory:
    @staticmethod
    def get_recommender(config: Dict[str, Any]) -> BaseRecommender:
        method = config.get("method")
        params = config.get("params", {})

        if method == "tfidf":
            return TFIDFRecommender(
                feature_matrix=params.get("feature_matrix")
            )
        # Add more elif blocks for additional content-based methods
        else:
            raise ValueError(f"Unsupported content-based filtering method: {method}")


# cold_start implementation
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import NearestNeighbors

class COLD_START:
    def __init__(self, method='content_based', n_neighbors=5):
        """
        Initialize cold start handling.
        
        Args:
            method (str): Cold start strategy ('content_based', 'popularity', 'hybrid')
            n_neighbors (int): Number of neighbors for similarity-based methods
        """
        self.method = method
        self.n_neighbors = n_neighbors
        self.scaler = StandardScaler()
        self.item_features = None
        self.popularity_scores = None
        self.knn = None

    def fit(self, item_features, interaction_matrix=None):
        """
        Fit the cold start handler.
        
        Args:
            item_features: Content features of items
            interaction_matrix: User-item interaction matrix (optional)
        """
        self.item_features = self.scaler.fit_transform(item_features)
        
        if interaction_matrix is not None:
            self.popularity_scores = np.sum(interaction_matrix, axis=0)
        
        if self.method in ['content_based', 'hybrid']:
            self.knn = NearestNeighbors(n_neighbors=self.n_neighbors)
            self.knn.fit(self.item_features)

    def recommend_for_new_user(self, user_profile=None):
        """
        Generate recommendations for a new user.
        
        Args:
            user_profile: User preferences or features (optional)
        """
        if self.method == 'content_based' and user_profile is not None:
            user_profile = self.scaler.transform([user_profile])
            _, indices = self.knn.kneighbors(user_profile)
            return indices[0]
            
        elif self.method == 'popularity':
            if self.popularity_scores is None:
                raise ValueError("Popularity scores not available")
            return np.argsort(self.popularity_scores)[-self.n_neighbors:]
            
        elif self.method == 'hybrid':
            content_based_scores = self._get_content_based_scores(user_profile)
            popularity_weights = self._normalize(self.popularity_scores)
            
            # Ensure both arrays have the same shape
            if content_based_scores.shape != popularity_weights.shape:
                # Align the shapes by selecting the top items based on content-based scores
                top_indices = np.argsort(content_based_scores)[-len(popularity_weights):]
                content_based_scores = content_based_scores[top_indices]
                popularity_weights = popularity_weights[top_indices]
            
            hybrid_scores = 0.7 * content_based_scores + 0.3 * popularity_weights
            return np.argsort(hybrid_scores)[-self.n_neighbors:]

    def _get_content_based_scores(self, user_profile):
        """Calculate content-based similarity scores."""
        if user_profile is None:
            return np.zeros(self.item_features.shape[0])
        user_profile = self.scaler.transform([user_profile])
        distances, _ = self.knn.kneighbors(user_profile)
        return self._normalize(1 / (1 + distances[0]))

    def _normalize(self, scores):
        """Normalize scores to [0, 1] range."""
        min_score = np.min(scores)
        max_score = np.max(scores)
        if min_score == max_score:
            return np.zeros_like(scores)  # Avoid division by zero
        return (scores - min_score) / (max_score - min_score)



# user_profiling implementation
import pandas as pd
from typing import List, Dict, Any, Optional

class UserProfilingRecommender:
    """
    A context-aware recommender system that builds and maintains user profiles for personalized recommendations.
    
    This recommender system creates and maintains detailed user profiles that capture user preferences,
    behavior patterns, and context-dependent interactions. It uses these profiles to generate
    personalized recommendations that adapt to different contexts.
    
    Attributes:
        user_profiles (Dict[int, Dict]): Detailed profiles for each user containing:
            - Preference vectors
            - Historical interactions
            - Context-dependent behavior patterns
            - Long-term and short-term interests
        context_preferences (Dict[int, Dict]): User preferences mapped to different contexts
        preference_evolution (Dict[int, List]): Temporal evolution of user preferences
    
    Methods:
        update_profile: Updates user profile with new interaction data
        analyze_context_preferences: Analyzes user preferences in different contexts
        detect_preference_shifts: Identifies changes in user preferences over time
        recommend: Generates personalized recommendations based on profile and context
    
    Example:
        >>> profiler = UserProfilingRecommender()
        >>> profiler.update_profile(user_id=123, interaction_data={...}, context={...})
        >>> recommendations = profiler.recommend(user_id=123, context={"location": "work"})
    
    Note:
        - Profiles are automatically updated with each user interaction
        - Supports both explicit (ratings) and implicit (behavior) feedback
        - Implements drift detection for evolving user preferences
    """
    def __init__(self, user_attributes: Optional[pd.DataFrame] = None):
        """
        Initialize the user profiling recommender with optional user attributes.

        Parameters:
        - user_attributes (pd.DataFrame, optional): DataFrame containing user information.
        """
        self.user_profiles: Dict[int, Dict[str, Any]] = {}
        self.user_attributes = user_attributes

    def fit(self, user_interactions: Dict[int, List[int]]):
        """
        Build user profiles based on interactions and user attributes.

        Parameters:
        - user_interactions (dict): Dictionary mapping user IDs to lists of interacted item IDs.
        """
        for user_id, items in user_interactions.items():
            profile = {}

            # Incorporate user attributes if available
            if self.user_attributes is not None:
                user_info = self.user_attributes[self.user_attributes['user_id'] == user_id]
                if not user_info.empty:
                    user_info = user_info.iloc[0]
                    for col in self.user_attributes.columns:
                        if col != 'user_id':
                            profile[col] = user_info[col]

            # Add interacted items
            profile['interacted_items'] = set(items)
            self.user_profiles[user_id] = profile

    def recommend(self, user_id: int, all_items: set, top_n: int = 10) -> List[int]:
        """
        Generate top-N item recommendations for a given user based on their profile.

        Parameters:
        - user_id (int): The ID of the user.
        - all_items (set): Set of all available item IDs.
        - top_n (int): The number of recommendations to generate.

        Returns:
        - List[int]: List of recommended item IDs.
        """
        if user_id not in self.user_profiles:
            return []

        interacted_items = self.user_profiles[user_id].get('interacted_items', set())
        recommendations = list(all_items - interacted_items)
        return recommendations[:top_n]
    


import pytest
import torch
from sage_conv import SAGEConv
from torch_geometric.data import Data

@pytest.fixture
def sample_data():
    # Create a simple graph with 4 nodes and 4 edges
    edge_index = torch.tensor([[0, 1, 2, 3], [1, 2, 3, 0]], dtype=torch.long)
    x = torch.tensor([[1, 0], [0, 1], [1, 1], [0, 0]], dtype=torch.float)
    return Data(x=x, edge_index=edge_index)

def test_sageconv_forward(sample_data):
    # Initialize SAGEConv layer
    conv = SAGEConv(in_channels=2, out_channels=2)

    # Perform a forward pass
    out = conv(sample_data.x, sample_data.edge_index)

    # Check output dimensions
    assert out.size() == (4, 2), "Output dimensions are incorrect"

def test_sageconv_with_different_aggregations(sample_data):
    # Test different aggregation methods
    for aggr in ['mean', 'max', 'sum']:
        conv = SAGEConv(in_channels=2, out_channels=2, aggr=aggr)
        out = conv(sample_data.x, sample_data.edge_index)
        assert out.size() == (4, 2), f"Output dimensions with {aggr} aggregation are incorrect"

def test_sageconv_with_normalization(sample_data):
    # Initialize SAGEConv layer with normalization
    conv = SAGEConv(in_channels=2, out_channels=2, normalize=True)

    # Perform a forward pass
    out = conv(sample_data.x, sample_data.edge_index)

    # Check output dimensions
    assert out.size() == (4, 2), "Output dimensions with normalization are incorrect"

def test_sageconv_without_root_weight(sample_data):
    # Initialize SAGEConv layer without root weight
    conv = SAGEConv(in_channels=2, out_channels=2, root_weight=False)

    # Perform a forward pass
    out = conv(sample_data.x, sample_data.edge_index)

    # Check output dimensions
    assert out.size() == (4, 2), "Output dimensions without root weight are incorrect"

def test_sageconv_with_projection(sample_data):
    # Initialize SAGEConv layer with projection
    conv = SAGEConv(in_channels=2, out_channels=2, project=True)

    # Perform a forward pass
    out = conv(sample_data.x, sample_data.edge_index)

    # Check output dimensions
    assert out.size() == (4, 2), "Output dimensions with projection are incorrect"

def test_sageconv_without_bias(sample_data):
    # Initialize SAGEConv layer without bias
    conv = SAGEConv(in_channels=2, out_channels=2, bias=False)

    # Perform a forward pass
    out = conv(sample_data.x, sample_data.edge_index)

    # Check output dimensions
    assert out.size() == (4, 2), "Output dimensions without bias are incorrect"

if __name__ == "__main__":
    pytest.main([__file__])



"""
This module provides functions to load, preprocess, and validate data from various file formats for use in a recommendation system.
## Author: Vishesh Yadav

Functions:
- register_format_detector(detector): Decorator to register a format detector.
- load_data(file_path): Loads data from a file, supporting multiple formats such as CSV, JSON, Excel, and more.
- load_xml(file_path): Loads data from an XML file into a DataFrame.
- load_hdf5(file_path): Loads data from an HDF5 file into a DataFrame.
- load_hdf4(file_path): Loads data from an HDF4 file into a DataFrame.
- load_avro(file_path): Loads data from an Avro file into a DataFrame.
- load_sqlite(file_path): Loads data from a SQLite database into a DataFrame.
- load_orc(file_path): Loads data from an ORC file into a DataFrame.
- load_sql_export(file_path): Placeholder for loading data from SQL Server or PostgreSQL exports.
- load_bson(file_path): Loads data from a BSON file (MongoDB export) into a DataFrame.
- load_latex(file_path): Loads data from a LaTeX table into a DataFrame.
- load_access_db(file_path): Loads data from an Access database into a DataFrame.
- detect_format(df): Detects the format of the dataset, such as interaction list or matrix.
- preprocess_data(df, data_format): Preprocesses data based on its detected format.
- validate_data(df): Validates the data for use in a recommendation system.
"""
import pandas as pd
import numpy as np
import json
import os
import warnings
import logging

from corerec.format_master.format_library import *

# Registry for format detection functions
FORMAT_DETECTORS = []

def register_format_detector(func):
    """Decorator to register a format detector."""
    FORMAT_DETECTORS.append(func)
    return func

def load_data(file_path):
    """Load data from various formats."""
    _, file_extension = os.path.splitext(file_path)
    file_extension = file_extension.lower()
    
    if file_extension == '.csv':
        # Specify dtypes to avoid dtype inference issues
        dtype_spec = {
            'isbn': 'object',
            'isbn13': 'object',
            'original_publication_year': 'float64',
            'average_rating': 'float64',
            'ratings_count': 'int64'
        }
        return pd.read_csv(file_path, dtype=dtype_spec)
    
    elif file_extension == '.tsv':
        return pd.read_csv(file_path, sep='\t')
    
    elif file_extension == '.json':
        with open(file_path, 'r') as f:
            return pd.DataFrame(json.load(f))
    
    elif file_extension in ['.xls', '.xlsx']:
        return pd.read_excel(file_path, sheet_name=None)  # Load all sheets as a dict of DataFrames
    
    elif file_extension == '.parquet' and 'pyarrow.parquet' in globals():
        return pd.read_parquet(file_path)
    
    elif file_extension == '.xml' and 'ET' in globals():
        return load_xml(file_path)
    
    elif file_extension == '.dat':
        return pd.read_csv(file_path, delimiter='::', engine='python', names=['user', 'item', 'rating', 'timestamp'])
    
    elif file_extension == '.yaml' and 'yaml' in globals():
        with open(file_path, 'r') as f:
            return pd.DataFrame(yaml.safe_load(f))
    
    elif file_extension == '.h5' and 'h5py' in globals():
        return load_hdf5(file_path)
    
    elif file_extension == '.hdf4' and 'h5py' in globals():
        return load_hdf4(file_path)
    
    elif file_extension == '.feather' and 'feather' in globals():
        return pd.read_feather(file_path)
    
    elif file_extension == '.avro' and 'avro' in globals():
        return load_avro(file_path)
    
    elif file_extension in ['.db', '.sqlite']:
        return load_sqlite(file_path)
    
    elif file_extension == '.tsv' and 'pandas.io.parsers' in globals():
        return pd.read_fwf(file_path)
    
    elif file_extension == '.dta':
        if 'pandas.io.parsers' in globals():
            return pd.read_stata(file_path)
        else:
            warnings.warn("pandas.io.parsers is not available. Stata files will not be supported.")
    
    elif file_extension == '.sav':
        if 'pandas.io.parsers' in globals():
            return pd.read_spss(file_path)
        else:
            warnings.warn("pandas.io.parsers is not available. SPSS files will not be supported.")
    
    elif file_extension == '.pkl':
        return pd.read_pickle(file_path)
    
    elif file_extension == '.msgpack' and 'msgpack' in globals():
        return pd.read_msgpack(file_path)
    
    elif file_extension == '.pb' and 'pystan' in globals():
        # Example Protocol Buffers loader (custom implementation needed)
        warnings.warn("Protocol Buffers loading is not implemented.")
        return pd.DataFrame()
    
    elif file_extension == '.orc' and 'pyorc' in globals():
        return load_orc(file_path)
    
    elif file_extension == '.fwf':
        return pd.read_fwf(file_path)
    
    elif file_extension == '.geojson':
        return pd.read_json(file_path)
    
    elif file_extension == '.toml' and 'toml' in globals():
        with open(file_path, 'r') as f:
            data = toml.load(f)
            return pd.DataFrame(data)
    
    elif file_extension == '.html':
        return pd.read_html(file_path)
    
    elif file_extension == '.sql' and 'sqlalchemy' in globals():
        return load_sql_export(file_path)
    
    elif file_extension == '.bson' and 'pymongo' in globals():
        return load_bson(file_path)
    
    elif file_extension in ['.tex'] and 'textract' in globals():
        return load_latex(file_path)
    
    elif file_extension in ['.md']:
        warnings.warn("Markdown files are not supported.")
        return pd.DataFrame()
    
    elif file_extension in ['.sas7bdat']:
        if 'pandas.io.parsers' in globals():
            return pd.read_sas(file_path, format='sas7bdat')
        else:
            warnings.warn("pandas.io.parsers is not available. SAS files will not be supported.")
    
    elif file_extension in ['.accdb', '.mdb']:
        if 'pyodbc' in globals():
            return load_access_db(file_path)
        else:
            warnings.warn("pyodbc is not available. Access DB files will not be supported.")
    
    elif file_extension in ['.tex']:
        if 'textract' in globals():
            return load_latex(file_path)
        else:
            warnings.warn("Textract is not available. LaTeX files will not be supported.")
    
    elif file_extension in ['.txt']:
        return pd.read_csv(file_path, delimiter='\t')
    
    elif file_extension in ['.dat']:
        return pd.read_csv(file_path, delimiter='::', engine='python', names=['user', 'item', 'rating', 'timestamp'])
    
    elif file_extension in ['.yaml']:
        with open(file_path, 'r') as f:
            return pd.DataFrame(yaml.safe_load(f))
    
    else:
        warnings.warn(f"Unsupported file format: {file_extension}")
        return pd.DataFrame()

def load_xml(file_path):
    """Load data from an XML file."""
    tree = ET.parse(file_path)
    root = tree.getroot()
    data = []
    for child in root:
        row = {}
        for subchild in child:
            row[subchild.tag] = subchild.text
        data.append(row)
    return pd.DataFrame(data)

def load_hdf5(file_path):
    """Load data from an HDF5 file."""
    with h5py.File(file_path, 'r') as f:
        data = {key: np.array(f[key]) for key in f.keys()}
    return pd.DataFrame(data)

def load_hdf4(file_path):
    """Load data from an HDF4 file."""
    try:
        import pyhdf.SD as SD
        hdf = SD.SD(file_path)
        data = {}
        for sds in hdf:
            data[sds.name()] = sds.get()
        return pd.DataFrame(data)
    except ImportError:
        warnings.warn("pyhdf is not available. HDF4 files will not be supported.")
        return pd.DataFrame()

def load_avro(file_path):
    """Load data from an Avro file."""
    try:
        with open(file_path, 'rb') as f:
            reader = avro.datafile.DataFileReader(f, avro.io.DatumReader())
            data = [record for record in reader]
            reader.close()
        return pd.DataFrame(data)
    except Exception as e:
        warnings.warn(f"Failed to load Avro file: {e}")
        return pd.DataFrame()

def load_sqlite(file_path):
    """Load data from a SQLite database."""
    try:
        conn = sqlite3.connect(file_path)
        query = "SELECT * FROM main_table"  # Replace with your table name
        df = pd.read_sql_query(query, conn)
        conn.close()
        return df
    except Exception as e:
        warnings.warn(f"Failed to load SQLite database: {e}")
        return pd.DataFrame()

def load_orc(file_path):
    """Load data from an ORC file."""
    try:
        import pyorc
        with open(file_path, 'rb') as f:
            reader = pyorc.Reader(f)
            columns = reader.schema.fields
            data = [dict(zip(columns, row)) for row in reader]
        return pd.DataFrame(data)
    except Exception as e:
        warnings.warn(f"Failed to load ORC file: {e}")
        return pd.DataFrame()

def load_sql_export(file_path):
    """Load data from SQL Server or PostgreSQL export."""
    try:
        # This is a placeholder. Implement actual SQL import as needed.
        warnings.warn("SQL export loading is not implemented.")
        return pd.DataFrame()
    except Exception as e:
        warnings.warn(f"Failed to load SQL export: {e}")
        return pd.DataFrame()

def load_bson(file_path):
    """Load data from a BSON file (MongoDB export)."""
    try:
        import bson
        with open(file_path, 'rb') as f:
            data = bson.decode_all(f.read())
        return pd.DataFrame(data)
    except Exception as e:
        warnings.warn(f"Failed to load BSON file: {e}")
        return pd.DataFrame()

def load_latex(file_path):
    """Load data from a LaTeX table."""
    try:
        text = textract.process(file_path).decode('utf-8')
        # This is a simplistic parser; you may need a more robust solution
        return pd.read_csv(pd.compat.StringIO(text), sep='&', engine='python').replace('\\', '', regex=True)
    except Exception as e:
        warnings.warn(f"Failed to load LaTeX file: {e}")
        return pd.DataFrame()

def load_access_db(file_path):
    """Load data from an Access database."""
    try:
        import pyodbc
        conn_str = (
            r'DRIVER={Microsoft Access Driver (*.mdb, *.accdb)};'
            f'DBQ={file_path};'
        )
        conn = pyodbc.connect(conn_str)
        query = "SELECT * FROM main_table"  # Replace with your table name
        df = pd.read_sql(query, conn)
        conn.close()
        return df
    except Exception as e:
        warnings.warn(f"Failed to load Access DB: {e}")
        return pd.DataFrame()

@register_format_detector
def detect_interaction_list(df):
    """Detect if the format is an interaction list."""
    if {'user', 'item', 'rating'}.issubset(df.columns):
        return 'interaction_list'
    return None

@register_format_detector
def detect_implicit_feedback(df):
    """Detect if the format is implicit feedback."""
    if {'user', 'item'}.issubset(df.columns) and 'rating' not in df.columns:
        return 'implicit_feedback'
    return None

@register_format_detector
def detect_interaction_matrix(df):
    """Detect if the format is an interaction matrix."""
    if df.index.name == 'user' and all(isinstance(col, str) for col in df.columns):
        return 'interaction_matrix'
    return None

@register_format_detector
def detect_numerical_matrix(df):
    """Detect if the format is a numerical matrix."""
    if df.dtypes.apply(lambda x: np.issubdtype(x, np.number)).all():
        return 'numerical_matrix'
    return None

@register_format_detector
def detect_categorical_matrix(df):
    """Detect if the format is a categorical matrix."""
    if df.dtypes.apply(lambda x: np.issubdtype(x, np.object_)).all():
        return 'categorical_matrix'
    return None

@register_format_detector
def detect_time_series(df):
    """Detect if the format is a time series."""
    if 'timestamp' in df.columns and pd.api.types.is_datetime64_any_dtype(df['timestamp']):
        return 'time_series'
    return None

@register_format_detector
def detect_geospatial(df):
    """Detect if the format contains geospatial data."""
    if {'latitude', 'longitude'}.issubset(df.columns):
        return 'geospatial'
    return None

@register_format_detector
def detect_user_user_matrix(df):
    """Detect if the format is a user-user matrix."""
    if df.index.name == 'user' and all(isinstance(col, str) for col in df.columns):
        return 'user_user_matrix'
    return None

@register_format_detector
def detect_user_item_matrix(df):
    """Detect if the format is a user-item matrix."""
    if 'user' in df.index.name and 'item' in df.columns:
        return 'user_item_matrix'
    return None

@register_format_detector
def detect_item_item_matrix(df):
    """Detect if the format is an item-item matrix."""
    if df.index.name == 'item' and all(isinstance(col, str) for col in df.columns):
        return 'item_item_matrix'
    return None

@register_format_detector
def detect_user_item_list(df):
    """Detect if the format is a user-item list."""
    if {'user', 'item'}.issubset(df.columns):
        return 'user_item_list'
    return None

@register_format_detector
def detect_user_item_rating_list(df):
    """Detect if the format is a user-item-rating list."""
    if {'user', 'item', 'rating'}.issubset(df.columns):
        return 'user_item_rating_list'
    return None

@register_format_detector
def detect_text_data(df):
    """Detect if the format contains text data."""
    if 'text' in df.columns or 'document' in df.columns:
        return 'text_data'
    return None

@register_format_detector
def detect_image_data(df):
    """Detect if the format contains image data."""
    if 'image_path' in df.columns or 'image_url' in df.columns:
        return 'image_data'
    return None

@register_format_detector
def detect_audio_data(df):
    """Detect if the format contains audio data."""
    if 'audio_path' in df.columns or 'audio_url' in df.columns:
        return 'audio_data'
    return None

@register_format_detector
def detect_video_data(df):
    """Detect if the format contains video data."""
    if 'video_path' in df.columns or 'video_url' in df.columns:
        return 'video_data'
    return None

@register_format_detector
def detect_explicit_feedback(df):
    """Detect if the format is explicit feedback."""
    if {'user', 'item', 'rating'}.issubset(df.columns):
        return 'explicit_feedback'
    return None

@register_format_detector
def detect_contextual_data(df):
    """Detect if the format contains contextual data."""
    if {'context', 'user', 'item'}.issubset(df.columns):
        return 'contextual_data'
    return None

@register_format_detector
def detect_session_data(df):
    """Detect if the format contains session data."""
    if 'session_id' in df.columns:
        return 'session_data'
    return None

@register_format_detector
def detect_demographic_data(df):
    """Detect if the format contains demographic data."""
    if {'age', 'gender', 'location'}.issubset(df.columns):
        return 'demographic_data'
    return None

@register_format_detector
def detect_social_network_data(df):
    """Detect if the format contains social network data."""
    if {'user', 'friend'}.issubset(df.columns):
        return 'social_network_data'
    return None

@register_format_detector
def detect_purchase_history(df):
    """Detect if the format contains purchase history."""
    if {'user', 'item', 'purchase_date'}.issubset(df.columns):
        return 'purchase_history'
    return None

@register_format_detector
def detect_browsing_history(df):
    """Detect if the format contains browsing history."""
    if {'user', 'page', 'timestamp'}.issubset(df.columns):
        return 'browsing_history'
    return None

@register_format_detector
def detect_clickstream_data(df):
    """Detect if the format contains clickstream data."""
    if {'user', 'click', 'timestamp'}.issubset(df.columns):
        return 'clickstream_data'
    return None

@register_format_detector
def detect_search_queries(df):
    """Detect if the format contains search queries."""
    if {'user', 'query', 'timestamp'}.issubset(df.columns):
        return 'search_queries'
    return None

@register_format_detector
def detect_user_profiles(df):
    """Detect if the format contains user profiles."""
    if {'user_id', 'age', 'gender', 'location'}.issubset(df.columns):
        return 'user_profiles'
    return None

@register_format_detector
def detect_item_metadata(df):
    """Detect if the format contains item metadata."""
    if {'item_id', 'description', 'category'}.issubset(df.columns):
        return 'item_metadata'
    return None

@register_format_detector
def detect_transaction_data(df):
    """Detect if the format contains transaction data."""
    if {'user_id', 'item_id', 'transaction_date'}.issubset(df.columns):
        return 'transaction_data'
    return None

@register_format_detector
def detect_ab_test_results(df):
    """Detect if the format contains A/B test results."""
    if {'test_id', 'variant', 'conversion_rate'}.issubset(df.columns):
        return 'ab_test_results'
    return None

@register_format_detector
def detect_user_engagement_metrics(df):
    """Detect if the format contains user engagement metrics."""
    if {'user_id', 'engagement_score', 'timestamp'}.issubset(df.columns):
        return 'user_engagement_metrics'
    return None

@register_format_detector
def detect_content_features(df):
    """Detect if the format contains content features."""
    if {'item_id', 'text_features', 'image_features'}.issubset(df.columns):
        return 'content_features'
    return None

@register_format_detector
def detect_collaborative_filtering_data(df):
    """Detect if the format contains collaborative filtering data."""
    if {'user_id', 'item_id', 'rating'}.issubset(df.columns):
        return 'collaborative_filtering_data'
    return None

@register_format_detector
def detect_implicit_feedback_data(df):
    """Detect if the format contains implicit feedback data."""
    if {'user_id', 'item_id', 'interaction_strength'}.issubset(df.columns):
        return 'implicit_feedback_data'
    return None

@register_format_detector
def detect_explicit_feedback_data(df):
    """Detect if the format contains explicit feedback data."""
    if {'user_id', 'item_id', 'rating'}.issubset(df.columns):
        return 'explicit_feedback_data'
    return None

@register_format_detector
def detect_time_series_data(df):
    """Detect if the format contains time series data."""
    if {'timestamp', 'value'}.issubset(df.columns):
        return 'time_series_data'
    return None

@register_format_detector
def detect_geospatial_data(df):
    """Detect if the format contains geospatial data."""
    if {'latitude', 'longitude', 'user_id'}.issubset(df.columns):
        return 'geospatial_data'
    return None

# Additional detectors to reach 100
@register_format_detector
def detect_user_item_interactions(df):
    """Detect if the format contains user-item interactions."""
    if {'user_id', 'item_id', 'interaction_type'}.issubset(df.columns):
        return 'user_item_interactions'
    return None

@register_format_detector
def detect_user_item_rating_interactions(df):
    """Detect if the format contains user-item-rating interactions."""
    if {'user_id', 'item_id', 'rating', 'interaction_type'}.issubset(df.columns):
        return 'user_item_rating_interactions'
    return None

@register_format_detector
def detect_user_activity_logs(df):
    """Detect if the format contains user activity logs."""
    if {'user_id', 'activity_type', 'timestamp'}.issubset(df.columns):
        return 'user_activity_logs'
    return None

@register_format_detector
def detect_item_view_logs(df):
    """Detect if the format contains item view logs."""
    if {'user_id', 'item_id', 'view_timestamp'}.issubset(df.columns):
        return 'item_view_logs'
    return None

@register_format_detector
def detect_user_feedback_logs(df):
    """Detect if the format contains user feedback logs."""
    if {'user_id', 'item_id', 'feedback', 'timestamp'}.issubset(df.columns):
        return 'user_feedback_logs'
    return None

@register_format_detector
def detect_user_preferences(df):
    """Detect if the format contains user preferences."""
    if {'user_id', 'preference_type', 'preference_value'}.issubset(df.columns):
        return 'user_preferences'
    return None

@register_format_detector
def detect_item_recommendations(df):
    """Detect if the format contains item recommendations."""
    if {'user_id', 'recommended_item_id', 'recommendation_score'}.issubset(df.columns):
        return 'item_recommendations'
    return None

@register_format_detector
def detect_user_item_similarity(df):
    """Detect if the format contains user-item similarity scores."""
    if {'user_id', 'item_id', 'similarity_score'}.issubset(df.columns):
        return 'user_item_similarity'
    return None

@register_format_detector
def detect_item_item_similarity(df):
    """Detect if the format contains item-item similarity scores."""
    if {'item_id_1', 'item_id_2', 'similarity_score'}.issubset(df.columns):
        return 'item_item_similarity'
    return None

@register_format_detector
def detect_user_item_rating_distribution(df):
    """Detect if the format contains user-item rating distributions."""
    if {'user_id', 'item_id', 'rating_distribution'}.issubset(df.columns):
        return 'user_item_rating_distribution'
    return None

@register_format_detector
def detect_user_item_interaction_counts(df):
    """Detect if the format contains user-item interaction counts."""
    if {'user_id', 'item_id', 'interaction_count'}.issubset(df.columns):
        return 'user_item_interaction_counts'
    return None

@register_format_detector
def detect_user_item_engagement(df):
    """Detect if the format contains user-item engagement metrics."""
    if {'user_id', 'item_id', 'engagement_score'}.issubset(df.columns):
        return 'user_item_engagement'
    return None

@register_format_detector
def detect_user_item_clicks(df):
    """Detect if the format contains user-item click data."""
    if {'user_id', 'item_id', 'click_count'}.issubset(df.columns):
        return 'user_item_clicks'
    return None

@register_format_detector
def detect_user_item_views(df):
    """Detect if the format contains user-item view data."""
    if {'user_id', 'item_id', 'view_count'}.issubset(df.columns):
        return 'user_item_views'
    return None

@register_format_detector
def detect_user_item_purchases(df):
    """Detect if the format contains user-item purchase data."""
    if {'user_id', 'item_id', 'purchase_count'}.issubset(df.columns):
        return 'user_item_purchases'
    return None

@register_format_detector
def detect_user_item_ratings(df):
    """Detect if the format is user-item ratings."""
    if {'user_id', 'item_id', 'rating'}.issubset(df.columns):
        return 'user_item_ratings'
    return None

@register_format_detector
def detect_user_item_feedback(df):
    """Detect if the format contains user-item feedback."""
    if {'user_id', 'item_id', 'feedback'}.issubset(df.columns):
        return 'user_item_feedback'
    return None

@register_format_detector
def detect_user_item_recommendations(df):
    """Detect if the format contains user-item recommendations."""
    if {'user_id', 'recommended_item_id', 'recommendation_score'}.issubset(df.columns):
        return 'user_item_recommendations'
    return None

@register_format_detector
def detect_user_item_interaction_types(df):
    """Detect if the format contains user-item interaction types."""
    if {'user_id', 'item_id', 'interaction_type'}.issubset(df.columns):
        return 'user_item_interaction_types'
    return None

@register_format_detector
def detect_user_item_contextual_data(df):
    """Detect if the format contains user-item contextual data."""
    if {'user_id', 'item_id', 'context_type', 'context_value'}.issubset(df.columns):
        return 'user_item_contextual_data'
    return None

@register_format_detector
def detect_user_item_session_data(df):
    """Detect if the format contains user-item session data."""
    if {'user_id', 'item_id', 'session_id', 'timestamp'}.issubset(df.columns):
        return 'user_item_session_data'
    return None

@register_format_detector
def detect_user_item_time_series(df):
    """Detect if the format contains user-item time series data."""
    if {'user_id', 'item_id', 'timestamp', 'value'}.issubset(df.columns):
        return 'user_item_time_series'
    return None

@register_format_detector
def detect_user_item_geospatial_data(df):
    """Detect if the format contains user-item geospatial data."""
    if {'user_id', 'item_id', 'latitude', 'longitude'}.issubset(df.columns):
        return 'user_item_geospatial_data'
    return None

@register_format_detector
def detect_user_item_demographics(df):
    """Detect if the format contains user-item demographic data."""
    if {'user_id', 'item_id', 'age', 'gender'}.issubset(df.columns):
        return 'user_item_demographics'
    return None

@register_format_detector
def detect_user_item_social_data(df):
    """Detect if the format contains user-item social data."""
    if {'user_id', 'item_id', 'friend_id'}.issubset(df.columns):
        return 'user_item_social_data'
    return None

@register_format_detector
def detect_user_item_purchase_history(df):
    """Detect if the format contains user-item purchase history."""
    if {'user_id', 'item_id', 'purchase_date'}.issubset(df.columns):
        return 'user_item_purchase_history'
    return None

@register_format_detector
def detect_user_item_browsing_history(df):
    """Detect if the format contains user-item browsing history."""
    if {'user_id', 'item_id', 'browsing_timestamp'}.issubset(df.columns):
        return 'user_item_browsing_history'
    return None

@register_format_detector
def detect_user_item_clickstream(df):
    """Detect if the format contains user-item clickstream data."""
    if {'user_id', 'item_id', 'click_timestamp'}.issubset(df.columns):
        return 'user_item_clickstream'
    return None

@register_format_detector
def detect_user_item_search_history(df):
    """Detect if the format contains user-item search history."""
    if {'user_id', 'item_id', 'search_query', 'search_timestamp'}.issubset(df.columns):
        return 'user_item_search_history'
    return None

@register_format_detector
def detect_user_item_engagement_metrics(df):
    """Detect if the format contains user-item engagement metrics."""
    if {'user_id', 'item_id', 'engagement_score'}.issubset(df.columns):
        return 'user_item_engagement_metrics'
    return None

@register_format_detector
def detect_user_item_recommendation_feedback(df):
    """Detect if the format contains user-item recommendation feedback."""
    if {'user_id', 'item_id', 'recommendation_feedback'}.issubset(df.columns):
        return 'user_item_recommendation_feedback'
    return None

@register_format_detector
def detect_user_item_rating_distribution(df):
    """Detect if the format contains user-item rating distributions."""
    if {'user_id', 'item_id', 'rating_distribution'}.issubset(df.columns):
        return 'user_item_rating_distribution'
    return None

@register_format_detector
def detect_user_item_interaction_distribution(df):
    """Detect if the format contains user-item interaction distributions."""
    if {'user_id', 'item_id', 'interaction_distribution'}.issubset(df.columns):
        return 'user_item_interaction_distribution'
    return None

@register_format_detector
def detect_user_item_engagement_distribution(df):
    """Detect if the format contains user-item engagement distributions."""
    if {'user_id', 'item_id', 'engagement_distribution'}.issubset(df.columns):
        return 'user_item_engagement_distribution'
    return None

@register_format_detector
def detect_user_item_click_distribution(df):
    """Detect if the format contains user-item click distributions."""
    if {'user_id', 'item_id', 'click_distribution'}.issubset(df.columns):
        return 'user_item_click_distribution'
    return None

@register_format_detector
def detect_user_item_view_distribution(df):
    """Detect if the format contains user-item view distributions."""
    if {'user_id', 'item_id', 'view_distribution'}.issubset(df.columns):
        return 'user_item_view_distribution'
    return None

@register_format_detector
def detect_user_item_purchase_distribution(df):
    """Detect if the format contains user-item purchase distributions."""
    if {'user_id', 'item_id', 'purchase_distribution'}.issubset(df.columns):
        return 'user_item_purchase_distribution'
    return None

@register_format_detector
def detect_user_item_rating_trends(df):
    """Detect if the format contains user-item rating trends."""
    if {'user_id', 'item_id', 'rating_trend'}.issubset(df.columns):
        return 'user_item_rating_trends'
    return None

@register_format_detector
def detect_user_item_engagement_trends(df):
    """Detect if the format contains user-item engagement trends."""
    if {'user_id', 'item_id', 'engagement_trend'}.issubset(df.columns):
        return 'user_item_engagement_trends'
    return None

@register_format_detector
def detect_user_item_click_trends(df):
    """Detect if the format contains user-item click trends."""
    if {'user_id', 'item_id', 'click_trend'}.issubset(df.columns):
        return 'user_item_click_trends'
    return None

@register_format_detector
def detect_user_item_view_trends(df):
    """Detect if the format contains user-item view trends."""
    if {'user_id', 'item_id', 'view_trend'}.issubset(df.columns):
        return 'user_item_view_trends'
    return None

@register_format_detector
def detect_user_item_purchase_trends(df):
    """Detect if the format contains user-item purchase trends."""
    if {'user_id', 'item_id', 'purchase_trend'}.issubset(df.columns):
        return 'user_item_purchase_trends'
    return None

@register_format_detector
def detect_user_item_feedback_trends(df):
    """Detect if the format contains user-item feedback trends."""
    if {'user_id', 'item_id', 'feedback_trend'}.issubset(df.columns):
        return 'user_item_feedback_trends'
    return None

@register_format_detector
def detect_user_item_recommendation_trends(df):
    """Detect if the format contains user-item recommendation trends."""
    if {'user_id', 'item_id', 'recommendation_trend'}.issubset(df.columns):
        return 'user_item_recommendation_trends'
    return None

@register_format_detector
def detect_user_item_engagement_metrics_trends(df):
    """Detect if the format contains user-item engagement metrics trends."""
    if {'user_id', 'item_id', 'engagement_metrics_trend'}.issubset(df.columns):
        return 'user_item_engagement_metrics_trends'
    return None

@register_format_detector
def detect_user_item_click_metrics_trends(df):
    """Detect if the format contains user-item click metrics trends."""
    if {'user_id', 'item_id', 'click_metrics_trend'}.issubset(df.columns):
        return 'user_item_click_metrics_trends'
    return None

@register_format_detector
def detect_user_item_view_metrics_trends(df):
    """Detect if the format contains user-item view metrics trends."""
    if {'user_id', 'item_id', 'view_metrics_trend'}.issubset(df.columns):
        return 'user_item_view_metrics_trends'
    return None

@register_format_detector
def detect_user_item_purchase_metrics_trends(df):
    """Detect if the format contains user-item purchase metrics trends."""
    if {'user_id', 'item_id', 'purchase_metrics_trend'}.issubset(df.columns):
        return 'user_item_purchase_metrics_trends'
    return None

@register_format_detector
def detect_user_item_rating_metrics_trends(df):
    """Detect if the format contains user-item rating metrics trends."""
    if {'user_id', 'item_id', 'rating_metrics_trend'}.issubset(df.columns):
        return 'user_item_rating_metrics_trends'
    return None

@register_format_detector
def detect_user_item_feedback_metrics_trends(df):
    """Detect if the format contains user-item feedback metrics trends."""
    if {'user_id', 'item_id', 'feedback_metrics_trend'}.issubset(df.columns):
        return 'user_item_feedback_metrics_trends'
    return None

@register_format_detector
def detect_user_item_recommendation_metrics_trends(df):
    """Detect if the format contains user-item recommendation metrics trends."""
    if {'user_id', 'item_id', 'recommendation_metrics_trend'}.issubset(df.columns):
        return 'user_item_recommendation_metrics_trends'
    return None

@register_format_detector
def detect_user_item_engagement_distribution_trends(df):
    """Detect if the format contains user-item engagement distribution trends."""
    if {'user_id', 'item_id', 'engagement_distribution_trend'}.issubset(df.columns):
        return 'user_item_engagement_distribution_trends'
    return None

@register_format_detector
def detect_user_item_click_distribution_trends(df):
    """Detect if the format contains user-item click distribution trends."""
    if {'user_id', 'item_id', 'click_distribution_trend'}.issubset(df.columns):
        return 'user_item_click_distribution_trends'
    return None

@register_format_detector
def detect_user_item_view_distribution_trends(df):
    """Detect if the format contains user-item view distribution trends."""
    if {'user_id', 'item_id', 'view_distribution_trend'}.issubset(df.columns):
        return 'user_item_view_distribution_trends'
    return None

@register_format_detector
def detect_user_item_purchase_distribution_trends(df):
    """Detect if the format contains user-item purchase distribution trends."""
    if {'user_id', 'item_id', 'purchase_distribution_trend'}.issubset(df.columns):
        return 'user_item_purchase_distribution_trends'
    return None

@register_format_detector
def detect_user_item_rating_distribution_trends(df):
    """Detect if the format contains user-item rating distribution trends."""
    if {'user_id', 'item_id', 'rating_distribution_trend'}.issubset(df.columns):
        return 'user_item_rating_distribution_trends'
    return None

@register_format_detector
def detect_user_item_engagement_trends_metrics(df):
    """Detect if the format contains user-item engagement trends metrics."""
    if {'user_id', 'item_id', 'engagement_trends_metrics'}.issubset(df.columns):
        return 'user_item_engagement_trends_metrics'
    return None

@register_format_detector
def detect_user_item_click_trends_metrics(df):
    """Detect if the format contains user-item click trends metrics."""
    if {'user_id', 'item_id', 'click_trends_metrics'}.issubset(df.columns):
        return 'user_item_click_trends_metrics'
    return None

@register_format_detector
def detect_user_item_view_trends_metrics(df):
    """Detect if the format contains user-item view trends metrics."""
    if {'user_id', 'item_id', 'view_trends_metrics'}.issubset(df.columns):
        return 'user_item_view_trends_metrics'
    return None

@register_format_detector
def detect_user_item_purchase_trends_metrics(df):
    """Detect if the format contains user-item purchase trends metrics."""
    if {'user_id', 'item_id', 'purchase_trends_metrics'}.issubset(df.columns):
        return 'user_item_purchase_trends_metrics'
    return None

@register_format_detector
def detect_user_item_rating_trends_metrics(df):
    """Detect if the format contains user-item rating trends metrics."""
    if {'user_id', 'item_id', 'rating_trends_metrics'}.issubset(df.columns):
        return 'user_item_rating_trends_metrics'
    return None

@register_format_detector
def detect_user_item_feedback_trends_metrics(df):
    """Detect if the format contains user-item feedback trends metrics."""
    if {'user_id', 'item_id', 'feedback_trends_metrics'}.issubset(df.columns):
        return 'user_item_feedback_trends_metrics'
    return None

@register_format_detector
def detect_user_item_recommendation_trends_metrics(df):
    """Detect if the format contains user-item recommendation trends metrics."""
    if {'user_id', 'item_id', 'recommendation_trends_metrics'}.issubset(df.columns):
        return 'user_item_recommendation_trends_metrics'
    return None

@register_format_detector
def detect_user_item_engagement_distribution_trends_metrics(df):
    """Detect if the format contains user-item engagement distribution trends metrics."""
    if {'user_id', 'item_id', 'engagement_distribution_trends_metrics'}.issubset(df.columns):
        return 'user_item_engagement_distribution_trends_metrics'
    return None

@register_format_detector
def detect_user_item_click_distribution_trends_metrics(df):
    """Detect if the format contains user-item click distribution trends metrics."""
    if {'user_id', 'item_id', 'click_distribution_trends_metrics'}.issubset(df.columns):
        return 'user_item_click_distribution_trends_metrics'
    return None

@register_format_detector
def detect_user_item_view_distribution_trends_metrics(df):
    """Detect if the format contains user-item view distribution trends metrics."""
    if {'user_id', 'item_id', 'view_distribution_trends_metrics'}.issubset(df.columns):
        return 'user_item_view_distribution_trends_metrics'
    return None

@register_format_detector
def detect_user_item_purchase_distribution_trends_metrics(df):
    """Detect if the format contains user-item purchase distribution trends metrics."""
    if {'user_id', 'item_id', 'purchase_distribution_trends_metrics'}.issubset(df.columns):
        return 'user_item_purchase_distribution_trends_metrics'
    return None

@register_format_detector
def detect_user_item_rating_distribution_trends_metrics(df):
    """Detect if the format contains user-item rating distribution trends metrics."""
    if {'user_id', 'item_id', 'rating_distribution_trends_metrics'}.issubset(df.columns):
        return 'user_item_rating_distribution_trends_metrics'
    return None

# Add more detectors as needed to reach 100

def detect_format(df):
    """Detect the format of the dataset."""
    for detector in FORMAT_DETECTORS:
        format_type = detector(df)
        if format_type:
            logging.info(f"Detected format: {format_type}")
            return format_type
    logging.warning("No matching format detected. Returning 'unknown'.")
    return 'unknown'

def preprocess_data(df, data_format):
    """Preprocess data based on its format."""
    if data_format == 'unknown':
        logging.error("Unknown data format. Preprocessing cannot proceed.")
        return None

    # Existing preprocessing logic...
    if data_format == 'interaction_list':
        if 'rating' in df.columns:
            df['rating'] = df['rating'] / df['rating'].max()
    elif data_format == 'implicit_feedback':
        df['interaction'] = 1
    elif data_format == 'interaction_matrix':
        df.fillna(0, inplace=True)
    elif data_format == 'numerical_matrix':
        df = (df - df.mean()) / df.std()
    elif data_format == 'categorical_matrix':
        df = pd.get_dummies(df)
    elif data_format == 'time_series':
        df['timestamp'] = pd.to_datetime(df['timestamp'])
        df.set_index('timestamp', inplace=True)
    elif data_format == 'geospatial':
        df['geo'] = df.apply(lambda row: (row['latitude'], row['longitude']), axis=1)
    elif data_format == 'book_metadata':
        df['average_rating'] = df['average_rating'].fillna(df['average_rating'].mean())
        df['title'] = df['title'].fillna('Unknown Title')
    elif data_format == 'user_user_matrix':
        df['user'] = df['user'].astype(str)
        df.set_index('user', inplace=True)
    else:
        logging.warning(f"No preprocessing steps defined for format: {data_format}")
    # Add more preprocessing steps as needed
    return df

def validate_data(df):
    """Validate the data for the recommendation system."""
    if df.empty:
        warnings.warn("The DataFrame is empty.")
    if df.isnull().values.any():
        warnings.warn("The DataFrame contains missing values.")
    # Add more validation checks as needed
    pass




import numpy as np

class Word2VecExtractor:
    def __init__(self, text_column, vector_size=100, window=5, min_count=1, epochs=10):
        self.text_column = text_column
        self.vector_size = vector_size
        self.window = window
        self.min_count = min_count
        self.epochs = epochs
        self.vocabulary = {}
        self.embeddings = {}

    def build_vocab(self, sentences):
        word_freq = defaultdict(int)
        for sentence in sentences:
            words = sentence.split('|')
            for word in words:
                word_freq[word] += 1
        self.vocabulary = {word: idx for idx, (word, freq) in enumerate(word_freq.items()) if freq >= self.min_count}
        self.embeddings = {word: np.random.rand(self.vector_size) for word in self.vocabulary}

    def train(self, sentences):
        # Simplified training: Average word vectors for each sentence
        for epoch in range(self.epochs):
            for sentence in sentences:
                words = sentence.split('|')
                valid_words = [word for word in words if word in self.vocabulary]
                if not valid_words:
                    continue
                for word in valid_words:
                    context = [w for w in valid_words if w != word]
                    if not context:
                        continue
                    for ctx_word in context:
                        # Update embeddings with a simple learning rule
                        self.embeddings[word] += 0.01 * (self.embeddings[ctx_word] - self.embeddings[word])

    def extract_features(self, data):
        sentences = data[self.text_column].tolist()
        self.build_vocab(sentences)
        self.train(sentences)

        feature_matrix = []
        for sentence in sentences:
            words = sentence.split('|')
            valid_words = [word for word in words if word in self.vocabulary]
            if not valid_words:
                feature_matrix.append(np.zeros(self.vector_size))
                continue
            vectors = [self.embeddings[word] for word in valid_words]
            feature_vector = np.mean(vectors, axis=0)
            feature_matrix.append(feature_vector.tolist())

        return feature_matrix


"""
Attention Mechanism Base Module

This module provides foundational classes and utilities for implementing attention-based 
unionized filtering mechanisms in recommendation systems. It includes implementations 
of various attention architectures like self-attention, multi-head attention, and 
transformer-based approaches.

Available Classes:
    - AttentionBasedUFBase: Base class for attention-based filtering
    - TransformerBasedUFBase: Base class for transformer architectures
    - SASRecBase: Base implementation of Self-Attentive Sequential Recommendation
"""

from .Transformer_based_uf_base import TransformerBasedUFBase as AM_TRANSFORMER
from .Attention_based_uf_base import AttentionBasedUFBase as AM_ATTENTION

from .a2svd import A2SVD as AM_A2SD
from .sasrec import SASRec as AM_SAS

__all__ = [
    "AM_TRANSFORMER",
    "AM_ATTENTION",
    "AM_A2SD",
    "AM_SAS",
]



# from .hybrid_collaborative import HYBRID_COLLABORATIVE as HYB_HYBRID_COLLABORATIVE
# from .attention_mechanisms import ATTENTION_MECHANISMS as HYB_ATTENTION_MECHANISMS
# from .ensemble_methods import ENSEMBLE_METHODS as HYB_ENSEMBLE_METHODS



# bayesian implementation BAYESIAN classifier (Naive Bayes)
import logging
from sklearn.naive_bayes import MultinomialNB
from sklearn.feature_extraction.text import CountVectorizer
from typing import List

# Configure logging
logger = logging.getLogger(__name__)

class BAYESIAN:
    def __init__(self):
        """
        Initialize the Bayesian classifier using Multinomial Naive Bayes.
        """
        self.vectorizer = CountVectorizer(stop_words='english')
        self.model = MultinomialNB()
        logger.info("Bayesian classifier initialized.")

    def fit(self, documents: List[str], labels: List[int]):
        """
        Fit the Bayesian classifier on the provided documents and labels.

        Parameters:
        - documents (List[str]): List of documents to train the model.
        - labels (List[int]): Corresponding labels for the documents.
        """
        logger.info("Fitting Bayesian classifier on documents.")
        count_matrix = self.vectorizer.fit_transform(documents)
        self.model.fit(count_matrix, labels)
        logger.info("Bayesian classifier training completed.")

    def predict(self, query: str) -> int:
        """
        Predict the label for a given query.

        Parameters:
        - query (str): The query text to classify.

        Returns:
        - int: Predicted label.
        """
        logger.info("Predicting label using Bayesian classifier.")
        query_vec = self.vectorizer.transform([query])
        prediction = self.model.predict(query_vec)[0]
        logger.info(f"Predicted label: {prediction}")
        return prediction

    def recommend(self, query: str, top_n: int = 10) -> List[int]:
        """
        Recommend items based on the Bayesian classifier's prediction.

        Parameters:
        - query (str): The query text for which to generate recommendations.
        - top_n (int): Number of top recommendations to return.

        Returns:
        - List[int]: List of recommended item indices.
        """
        logger.info("Generating recommendations using Bayesian classifier.")
        predicted_label = self.predict(query)
        # Example: Recommend items with the same label
        # This requires access to labeled items; here we return an empty list as a placeholder
        recommendations = []  # Implement logic based on the application
        logger.info(f"Top {top_n} recommendations generated using Bayesian classifier.")
        return recommendations[:top_n]



# interactive_filtering implementation

import logging
from typing import Dict, List, Any

# Configure logging
logger = logging.getLogger(__name__)

class InteractiveFilteringRecommender:
    def __init__(self, base_recommender: Any):
        """
        Initialize the InteractiveFilteringRecommender with a base recommender.

        Parameters:
        - base_recommender (Any): An instance of a base recommender (e.g., LSA, LDA).
        """
        self.base_recommender = base_recommender
        # Stores user feedback in the format {user_id: {item_id: feedback_score}}
        self.user_feedback: Dict[int, Dict[int, float]] = {}
        logger.info("InteractiveFilteringRecommender initialized with base recommender.")

    def collect_feedback(self, user_id: int, item_id: int, feedback_score: float):
        """
        Collect feedback from the user for a specific item.

        Parameters:
        - user_id (int): The ID of the user.
        - item_id (int): The ID of the item.
        - feedback_score (float): The feedback score (e.g., 1.0 for positive, -1.0 for negative).
        """
        if user_id not in self.user_feedback:
            self.user_feedback[user_id] = {}
        self.user_feedback[user_id][item_id] = feedback_score
        logger.info(f"Collected feedback from user {user_id} for item {item_id}: {feedback_score}")

        # Update the base recommender based on feedback
        self.update_recommender(user_id, item_id, feedback_score)

    def update_recommender(self, user_id: int, item_id: int, feedback_score: float):
        """
        Update the base recommender system based on user feedback.

        Parameters:
        - user_id (int): The ID of the user.
        - item_id (int): The ID of the item.
        - feedback_score (float): The feedback score.
        """
        try:
            # Example: Adjust user profile or model based on feedback
            # This is a placeholder for actual implementation
            logger.info(f"Updating base recommender for user {user_id} based on feedback.")
            if hasattr(self.base_recommender, 'update_user_profile'):
                self.base_recommender.update_user_profile(user_id, item_id, feedback_score)
            else:
                logger.warning("Base recommender does not support updating user profiles.")
        except Exception as e:
            logger.error(f"Error updating recommender based on feedback: {e}")

    def recommend(self, user_id: int, query: str, top_n: int = 10) -> List[int]:
        """
        Generate top-N item recommendations for a user, considering their feedback.

        Parameters:
        - user_id (int): The ID of the user.
        - query (str): The query text for generating recommendations.
        - top_n (int): Number of top recommendations to return.

        Returns:
        - List[int]: List of recommended item IDs.
        """
        logger.info(f"Generating recommendations for user {user_id} with query '{query}' using InteractiveFilteringRecommender.")
        # Get base recommendations
        base_recommendations = self.base_recommender.recommend(user_id, query, top_n=top_n * 2)

        if user_id in self.user_feedback:
            # Filter out items with negative feedback
            filtered_recommendations = [
                item_id for item_id in base_recommendations
                if self.user_feedback[user_id].get(item_id, 0) >= 0
            ]
            logger.info(f"Filtered recommendations for user {user_id}: {filtered_recommendations}")
            return filtered_recommendations[:top_n]
        else:
            return base_recommendations[:top_n]



# feature_extraction implementation
import logging
from sklearn.feature_extraction.text import TfidfVectorizer
from nltk.stem import WordNetLemmatizer
import nltk
from typing import List, Any

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Download NLTK resources if not already present
nltk.download('punkt', quiet=True)
nltk.download('wordnet', quiet=True)

class FeatureExtraction:
    def __init__(self, max_features=5000):
        """
        Initializes the FeatureExtraction with a TF-IDF vectorizer.

        Parameters:
        - max_features (int): The maximum number of features (vocabulary size).
        """
        self.max_features = max_features
        self.lemmatizer = WordNetLemmatizer()
        self.vectorizer = TfidfVectorizer(
            max_features=self.max_features,
            stop_words='english',  # Use built-in stop words
            tokenizer=self.tokenize
        )
        logger.info(f"FeatureExtraction initialized with max_features={self.max_features}.")

    def tokenize(self, text: str) -> List[str]:
        """
        Tokenizes and lemmatizes the input text.

        Parameters:
        - text (str): The text to tokenize.

        Returns:
        - list: A list of processed tokens.
        """
        tokens = nltk.word_tokenize(text.lower())
        lemmatized = [
            self.lemmatizer.lemmatize(token)
            for token in tokens
            if token.isalpha()
        ]
        logger.debug(f"Tokenized text: {lemmatized}")
        return lemmatized

    def fit_transform(self, documents: List[str]):
        """
        Fits the TF-IDF vectorizer on the documents and transforms them into feature vectors.

        Parameters:
        - documents (list of str): The list of documents to process.

        Returns:
        - sparse matrix: The TF-IDF feature matrix.
        """
        logger.info("Fitting and transforming documents into TF-IDF features.")
        return self.vectorizer.fit_transform(documents)

    def transform(self, documents: List[str]) -> Any:
        """
        Transforms the documents into TF-IDF feature vectors using the already fitted vectorizer.

        Parameters:
        - documents (list of str): The list of documents to transform.

        Returns:
        - sparse matrix: The TF-IDF feature matrix.
        """
        logger.info("Transforming documents into LSA latent space.")
        tfidf_matrix = self.vectorizer.transform(documents)  # Use transform, not fit
        return self.lsa_model.transform(tfidf_matrix)

    def get_feature_names(self) -> List[str]:
        """
        Retrieves the feature names (vocabulary) from the vectorizer.

        Returns:
        - list: A list of feature names.
        """
        return self.vectorizer.get_feature_names_out()



# corerec/engines/unionizedFilterEngine/lightgcn.py
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import List, Dict, Tuple, Optional
import pickle
import os
from scipy.sparse import csr_matrix

from ..base_recommender import BaseRecommender


class LightGCN(BaseRecommender):
    """
    LightGCN: Simplifying and Powering Graph Convolution Network for Recommendation
    
    This implementation follows the paper:
    "LightGCN: Simplifying and Powering Graph Convolution Network for Recommendation"
    by Xiangnan He, Kuan Deng, Xiang Wang, Yan Li, Yongdong Zhang, Meng Wang
    """
    
    def __init__(self, 
                 n_factors: int = 64, 
                 n_layers: int = 3,
                 learning_rate: float = 0.001,
                 regularization: float = 1e-5,
                 batch_size: int = 1024,
                 epochs: int = 100,
                 device: str = 'cuda' if torch.cuda.is_available() else 'cpu',
                 dropout: float = 0.0,
                 early_stopping_patience: int = 10,
                 verbose: bool = True):
        """
        Initialize LightGCN model
        
        Args:
            n_factors: Embedding size
            n_layers: Number of graph convolution layers
            learning_rate: Learning rate for optimizer
            regularization: L2 regularization coefficient
            batch_size: Training batch size
            epochs: Number of training epochs
            device: Computing device ('cpu' or 'cuda')
            dropout: Dropout rate
            early_stopping_patience: Number of epochs to wait before early stopping
            verbose: Whether to print training progress
        """
        super().__init__()
        self.n_factors = n_factors
        self.n_layers = n_layers
        self.learning_rate = learning_rate
        self.regularization = regularization
        self.batch_size = batch_size
        self.epochs = epochs
        self.device = device
        self.dropout = dropout
        self.early_stopping_patience = early_stopping_patience
        self.verbose = verbose
        
        # These will be initialized during fitting
        self.n_users = None
        self.n_items = None
        self.user_embedding = None
        self.item_embedding = None
        self.model = None
        self.optimizer = None
        
        # For mapping between original IDs and internal indices
        self.user_id_map = {}
        self.item_id_map = {}
        self.reverse_user_map = {}
        self.reverse_item_map = {}
        
        # For storing user interactions
        self.user_interactions = {}
        
    def _create_mappings(self, user_ids: List[int], item_ids: List[int]) -> None:
        """
        Create mappings between original IDs and internal indices
        
        Args:
            user_ids: List of user IDs
            item_ids: List of item IDs
        """
        unique_user_ids = sorted(set(user_ids))
        unique_item_ids = sorted(set(item_ids))
        
        self.user_id_map = {user_id: idx for idx, user_id in enumerate(unique_user_ids)}
        self.item_id_map = {item_id: idx for idx, item_id in enumerate(unique_item_ids)}
        
        self.reverse_user_map = {idx: user_id for user_id, idx in self.user_id_map.items()}
        self.reverse_item_map = {idx: item_id for item_id, idx in self.item_id_map.items()}
        
        self.n_users = len(unique_user_ids)
        self.n_items = len(unique_item_ids)
        
    def _build_model(self) -> None:
        """
        Build the LightGCN model
        """
        self.model = LightGCNModel(
            n_users=self.n_users,
            n_items=self.n_items,
            n_factors=self.n_factors,
            n_layers=self.n_layers,
            dropout=self.dropout
        ).to(self.device)
        
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), 
            lr=self.learning_rate, 
            weight_decay=self.regularization
        )
        
    def _create_adjacency_matrix(self, interaction_matrix: csr_matrix) -> torch.Tensor:
        """
        Create normalized adjacency matrix for graph convolution
        
        Args:
            interaction_matrix: User-item interaction matrix
            
        Returns:
            Normalized adjacency matrix as a sparse tensor
        """
        # Convert to PyTorch sparse tensor
        indices = interaction_matrix.nonzero()
        values = interaction_matrix.data
        
        # Create indices for the sparse tensor
        user_indices = torch.LongTensor(indices[0])
        item_indices = torch.LongTensor(indices[1])
        
        # Create edge index for the graph
        edge_index = torch.stack([
            torch.cat([user_indices, item_indices + self.n_users]),
            torch.cat([item_indices + self.n_users, user_indices])
        ])
        
        # Create adjacency matrix
        adj = torch.sparse.FloatTensor(
            edge_index,
            torch.ones(edge_index.size(1)),
            torch.Size([self.n_users + self.n_items, self.n_users + self.n_items])
        ).to(self.device)
        
        # Calculate degree matrix
        rowsum = torch.sparse.sum(adj, dim=1).to_dense()
        d_inv_sqrt = torch.pow(rowsum, -0.5)
        d_inv_sqrt[torch.isinf(d_inv_sqrt)] = 0.
        d_mat_inv_sqrt = torch.diag(d_inv_sqrt)
        
        # Calculate normalized adjacency matrix: D^(-0.5) * A * D^(-0.5)
        norm_adj = torch.sparse.mm(torch.sparse.mm(d_mat_inv_sqrt, adj), d_mat_inv_sqrt)
        
        return norm_adj
    
    def _store_user_interactions(self, user_ids: List[int], item_ids: List[int]) -> None:
        """
        Store user interactions for recommendation filtering
        
        Args:
            user_ids: List of user IDs
            item_ids: List of item IDs
        """
        self.user_interactions = {}
        for user_id, item_id in zip(user_ids, item_ids):
            user_idx = self.user_id_map.get(user_id)
            item_idx = self.item_id_map.get(item_id)
            
            if user_idx is not None and item_idx is not None:
                if user_idx not in self.user_interactions:
                    self.user_interactions[user_idx] = set()
                self.user_interactions[user_idx].add(item_idx)
    
    def _sample_negative_items(self, user_idx: int, n_neg: int = 1) -> List[int]:
        """
        Sample negative items for a user
        
        Args:
            user_idx: User index
            n_neg: Number of negative items to sample
            
        Returns:
            List of negative item indices
        """
        pos_items = self.user_interactions.get(user_idx, set())
        neg_items = []
        
        while len(neg_items) < n_neg:
            neg_item = np.random.randint(0, self.n_items)
            if neg_item not in pos_items and neg_item not in neg_items:
                neg_items.append(neg_item)
                
        return neg_items
    
    def _create_bpr_loss(self, users: torch.Tensor, pos_items: torch.Tensor, neg_items: torch.Tensor) -> torch.Tensor:
        """
        Create BPR loss for training
        
        Args:
            users: User indices
            pos_items: Positive item indices
            neg_items: Negative item indices
            
        Returns:
            BPR loss
        """
        # Get embeddings
        user_emb, item_emb = self.model()
        
        # Get user embeddings
        user_embeddings = user_emb[users]
        
        # Get positive and negative item embeddings
        pos_item_embeddings = item_emb[pos_items]
        neg_item_embeddings = item_emb[neg_items]
        
        # Calculate scores
        pos_scores = torch.sum(user_embeddings * pos_item_embeddings, dim=1)
        neg_scores = torch.sum(user_embeddings * neg_item_embeddings, dim=1)
        
        # Calculate BPR loss
        loss = -torch.mean(torch.log(torch.sigmoid(pos_scores - neg_scores)))
        
        # Add L2 regularization
        l2_loss = self.regularization * (
            torch.norm(user_embeddings) ** 2 + 
            torch.norm(pos_item_embeddings) ** 2 + 
            torch.norm(neg_item_embeddings) ** 2
        ) / len(users)
        
        return loss + l2_loss
    
    def fit(self, interaction_matrix: csr_matrix, user_ids: List[int], item_ids: List[int]) -> None:
        """
        Fit the LightGCN model
        
        Args:
            interaction_matrix: User-item interaction matrix
            user_ids: List of user IDs
            item_ids: List of item IDs
        """
        # Create mappings
        self._create_mappings(user_ids, item_ids)
        
        # Store user interactions
        self._store_user_interactions(user_ids, item_ids)
        
        # Build model
        self._build_model()
        
        # Create adjacency matrix
        norm_adj = self._create_adjacency_matrix(interaction_matrix)
        self.model.set_adj_matrix(norm_adj)
        
        # Training loop
        best_loss = float('inf')
        patience_counter = 0
        
        for epoch in range(self.epochs):
            self.model.train()
            
            # Sample training data
            users = []
            pos_items = []
            neg_items = []
            
            for user_idx in range(self.n_users):
                if user_idx in self.user_interactions:
                    for pos_item in self.user_interactions[user_idx]:
                        users.append(user_idx)
                        pos_items.append(pos_item)
                        neg_items.append(self._sample_negative_items(user_idx)[0])
            
            # Create batches
            n_batches = len(users) // self.batch_size + (1 if len(users) % self.batch_size != 0 else 0)
            
            total_loss = 0
            for batch_idx in range(n_batches):
                start_idx = batch_idx * self.batch_size
                end_idx = min((batch_idx + 1) * self.batch_size, len(users))
                
                batch_users = torch.LongTensor(users[start_idx:end_idx]).to(self.device)
                batch_pos_items = torch.LongTensor(pos_items[start_idx:end_idx]).to(self.device)
                batch_neg_items = torch.LongTensor(neg_items[start_idx:end_idx]).to(self.device)
                
                self.optimizer.zero_grad()
                loss = self._create_bpr_loss(batch_users, batch_pos_items, batch_neg_items)
                loss.backward()
                self.optimizer.step()
                
                total_loss += loss.item()
            
            avg_loss = total_loss / n_batches
            
            if self.verbose and (epoch + 1) % 10 == 0:
                print(f"Epoch {epoch+1}/{self.epochs}, Loss: {avg_loss:.4f}")
            
            # Early stopping
            if avg_loss < best_loss:
                best_loss = avg_loss
                patience_counter = 0
            else:
                patience_counter += 1
                
            if patience_counter >= self.early_stopping_patience:
                if self.verbose:
                    print(f"Early stopping at epoch {epoch+1}")
                break
        
        # Get final embeddings
        self.model.eval()
        self.user_embedding, self.item_embedding = self.model()
    
    def recommend(self, user_id: int, top_n: int = 10, exclude_seen: bool = True) -> List[int]:
        """
        Recommend items for a user
        
        Args:
            user_id: User ID
            top_n: Number of recommendations to return
            exclude_seen: Whether to exclude seen items
            
        Returns:
            List of recommended item IDs
        """
        if self.user_embedding is None or self.item_embedding is None:
            raise ValueError("Model has not been trained yet")
        
        if user_id not in self.user_id_map:
            return []
        
        user_idx = self.user_id_map[user_id]
        user_emb = self.user_embedding[user_idx].detach()
        
        # Calculate scores for all items
        scores = torch.matmul(user_emb, self.item_embedding.t()).detach().cpu().numpy()
        
        # Exclude seen items if needed
        if exclude_seen and user_idx in self.user_interactions:
            seen_items = self.user_interactions[user_idx]
            for item_idx in seen_items:
                scores[item_idx] = -np.inf
        
        # Get top-N items
        top_item_indices = np.argsort(-scores)[:top_n]
        
        # Convert indices back to original IDs
        top_items = [self.reverse_item_map[idx] for idx in top_item_indices]
        
        return top_items
    
    def predict(self, user_id: int, item_id: int) -> float:
        """
        Predict rating for a user-item pair
        
        Args:
            user_id: User ID
            item_id: Item ID
            
        Returns:
            Predicted rating
        """
        if self.user_embedding is None or self.item_embedding is None:
            raise ValueError("Model has not been trained yet")
        
        if user_id not in self.user_id_map or item_id not in self.item_id_map:
            return 0.0
        
        user_idx = self.user_id_map[user_id]
        item_idx = self.item_id_map[item_id]
        
        user_emb = self.user_embedding[user_idx]
        item_emb = self.item_embedding[item_idx]
        
        score = torch.dot(user_emb, item_emb).detach().cpu().numpy()
        
        return float(score)
    
    def save_model(self, filepath: str) -> None:
        """
        Save model to file
        
        Args:
            filepath: Path to save the model
        """
        model_data = {
            'n_factors': self.n_factors,
            'n_layers': self.n_layers,
            'learning_rate': self.learning_rate,
            'regularization': self.regularization,
            'batch_size': self.batch_size,
            'epochs': self.epochs,
            'device': self.device,
            'dropout': self.dropout,
            'early_stopping_patience': self.early_stopping_patience,
            'verbose': self.verbose,
            'n_users': self.n_users,
            'n_items': self.n_items,
            'user_id_map': self.user_id_map,
            'item_id_map': self.item_id_map,
            'reverse_user_map': self.reverse_user_map,
            'reverse_item_map': self.reverse_item_map,
            'user_interactions': self.user_interactions,
            'user_embedding': self.user_embedding.detach().cpu().numpy() if self.user_embedding is not None else None,
            'item_embedding': self.item_embedding.detach().cpu().numpy() if self.item_embedding is not None else None
        }
        
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        with open(filepath, 'wb') as f:
            pickle.dump(model_data, f)
    
    @classmethod
    def load_model(cls, filepath: str) -> 'LightGCN':
        """
        Load model from file
        
        Args:
            filepath: Path to load the model from
            
        Returns:
            Loaded LightGCN model
        """
        with open(filepath, 'rb') as f:
            model_data = pickle.load(f)
        
        instance = cls(
            n_factors=model_data['n_factors'],
            n_layers=model_data['n_layers'],
            learning_rate=model_data['learning_rate'],
            regularization=model_data['regularization'],
            batch_size=model_data['batch_size'],
            epochs=model_data['epochs'],
            device=model_data['device'],
            dropout=model_data['dropout'],
            early_stopping_patience=model_data['early_stopping_patience'],
            verbose=model_data['verbose']
        )
        
        instance.n_users = model_data['n_users']
        instance.n_items = model_data['n_items']
        instance.user_id_map = model_data['user_id_map']
        instance.item_id_map = model_data['item_id_map']
        instance.reverse_user_map = model_data['reverse_user_map']
        instance.reverse_item_map = model_data['reverse_item_map']
        instance.user_interactions = model_data['user_interactions']
        
        if model_data['user_embedding'] is not None and model_data['item_embedding'] is not None:
            instance.user_embedding = torch.tensor(model_data['user_embedding']).to(instance.device)
            instance.item_embedding = torch.tensor(model_data['item_embedding']).to(instance.device)
        
        return instance


class LightGCNModel(nn.Module):
    """
    LightGCN Model implementation
    """
    
    def __init__(self, n_users: int, n_items: int, n_factors: int, n_layers: int, dropout: float = 0.0):
        """
        Initialize LightGCN model
        
        Args:
            n_users: Number of users
            n_items: Number of items
            n_factors: Embedding size
            n_layers: Number of graph convolution layers
            dropout: Dropout rate
        """
        super(LightGCNModel, self).__init__()
        
        self.n_users = n_users
        self.n_items = n_items
        self.n_factors = n_factors
        self.n_layers = n_layers
        self.dropout = dropout
        
        # Initialize embeddings
        self.user_embedding = nn.Embedding(n_users, n_factors)
        self.item_embedding = nn.Embedding(n_items, n_factors)
        
        # Initialize weights
        nn.init.normal_(self.user_embedding.weight, std=0.1)
        nn.init.normal_(self.item_embedding.weight, std=0.1)
        
        # Adjacency matrix
        self.adj_matrix = None
    
    def set_adj_matrix(self, adj_matrix: torch.Tensor) -> None:
        """
        Set adjacency matrix for graph convolution
        
        Args:
            adj_matrix: Normalized adjacency matrix
        """
        self.adj_matrix = adj_matrix
    
    def forward(self) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Forward pass
        
        Returns:
            Tuple of user and item embeddings
        """
        if self.adj_matrix is None:
            raise ValueError("Adjacency matrix has not been set")
        
        # Get initial embeddings
        user_embeddings = self.user_embedding.weight
        item_embeddings = self.item_embedding.weight
        
        # Concatenate user and item embeddings
        all_embeddings = torch.cat([user_embeddings, item_embeddings], dim=0)
        
        # List to store embeddings from each layer
        embeddings_list = [all_embeddings]
        
        # Graph convolution
        for _ in range(self.n_layers):
            # Dropout if needed
            if self.dropout > 0 and self.training:
                all_embeddings = F.dropout(all_embeddings, p=self.dropout)
            
            # Perform graph convolution: A * H
            all_embeddings = torch.sparse.mm(self.adj_matrix, all_embeddings)
            
            # Add to list
            embeddings_list.append(all_embeddings)
        
        # Sum embeddings from all layers
        all_embeddings = torch.stack(embeddings_list, dim=1)
        all_embeddings = torch.mean(all_embeddings, dim=1)
        
        # Split user and item embeddings
        user_embeddings, item_embeddings = torch.split(all_embeddings, [self.n_users, self.n_items])
        
        return user_embeddings, item_embeddings


# mypy: allow-untyped-defs

import copyreg
import os.path as _osp
import weakref

import torch
from torch.utils import (
    backcompat as backcompat,
    collect_env as collect_env,
    data as data,
    hooks as hooks,
)
# from torch.utils.backend_registration import (
#     generate_methods_for_privateuse1_backend,
#     rename_privateuse1_backend,
# )
from torch.utils.cpp_backtrace import get_cpp_backtrace
from torch.utils.throughput_benchmark import ThroughputBenchmark


def set_module(obj, mod):
    """
    Set the module attribute on a python object for a given object for nicer printing
    """
    if not isinstance(mod, str):
        raise TypeError("The mod argument should be a string")
    obj.__module__ = mod


# If you need to check for a deployment environment, consider using an alternative approach
# For example, you might use environment variables or configuration files
# if torch._running_with_deploy():
#     # Handle the deploy-specific logic here
#     pass

def swap_tensors(t1, t2):
    """
    This function swaps the content of the two Tensor objects.
    At a high level, this will make t1 have the content of t2 while preserving
    its identity.

    This will not work if t1 and t2 have different slots.
    """
    # Ensure there are no weakrefs
    if weakref.getweakrefs(t1):
        raise RuntimeError("Cannot swap t1 because it has weakref associated with it")
    if weakref.getweakrefs(t2):
        raise RuntimeError("Cannot swap t2 because it has weakref associated with it")
    t1_slots = set(copyreg._slotnames(t1.__class__))  # type: ignore[attr-defined]
    t2_slots = set(copyreg._slotnames(t2.__class__))  # type: ignore[attr-defined]
    if t1_slots != t2_slots:
        raise RuntimeError("Cannot swap t1 and t2 if they have different slots")

    def swap_attr(name):
        tmp = getattr(t1, name)
        setattr(t1, name, (getattr(t2, name)))
        setattr(t2, name, tmp)

    def error_pre_hook(grad_outputs):
        raise RuntimeError(
            "Trying to execute AccumulateGrad node that was poisoned by swap_tensors "
            "this can happen when you try to run backward on a tensor that was swapped. "
            "For a module m with `torch.__future__.set_swap_module_params_on_conversion(True)` "
            "you should not change the device or dtype of the module (e.g. `m.cpu()` or `m.half()`) "
            "between running forward and backward. To resolve this, please only change the "
            "device/dtype before running forward (or after both forward and backward)."
        )

    def check_use_count(t, name="t1"):
        use_count = t._use_count()
        error_str = (
            f"Expected use_count of {name} to be 1 or 2 with an AccumulateGrad node but got {use_count} "
            f"make sure you are not holding references to the tensor in other places."
        )
        if use_count > 1:
            if use_count == 2 and t.is_leaf:
                accum_grad_node = torch.autograd.graph.get_gradient_edge(t).node
                # Make sure that the accumulate_grad node was not lazy_init-ed by get_gradient_edge
                if t._use_count() == 2:
                    accum_grad_node.register_prehook(error_pre_hook)
                else:
                    raise RuntimeError(error_str)
            else:
                raise RuntimeError(error_str)

    check_use_count(t1, "t1")
    check_use_count(t2, "t2")

    # Swap the types
    # Note that this will fail if there are mismatched slots
    swap_attr("__class__")

    # Swap the dynamic attributes
    swap_attr("__dict__")

    # Swap the slots
    for slot in t1_slots:
        if hasattr(t1, slot) and hasattr(t2, slot):
            swap_attr(slot)
        elif hasattr(t1, slot):
            setattr(t2, slot, (getattr(t1, slot)))
            delattr(t1, slot)
        elif hasattr(t2, slot):
            setattr(t1, slot, (getattr(t2, slot)))
            delattr(t2, slot)

    # Swap the at::Tensor they point to
    torch._C._swap_tensor_impl(t1, t2)

# Define the is_compiling function
def is_compiling() -> bool:
    # Implement the logic for determining if the code is compiling
    # This is just a placeholder implementation
    return False

# Export the function
__all__ = ['is_compiling']
